/** @file config.cpp

	Skelton for retropc emulator

	@author Takeda.Toshiya
	@date   2006.08.18 -

	@note
	Modified for BML3MK5 by Sasaji at 2011.06.17

	@brief [ config ]
*/

#include <stdlib.h>
#include "SimpleIni.h"
#include "config.h"
#include "fileio.h"
#include "emu.h"

extern EMU *emu;

config_t config;

static bool load_cfg_file();
static void save_cfg_file();
static bool load_ini_file();
static void save_ini_file();

//
// definition for ini file
// 
static CSimpleIni *ini = NULL;
#define SECTION_MAIN	_T("")
#define SECTION_CONTROL	_T("control")
#define SECTION_TAPE	_T("tape")
#define SECTION_FDD		_T("fdd")
#define SECTION_SCREEN	_T("screen")
#define SECTION_SOUND	_T("sound")
#define SECTION_PRINTER	_T("printer")
#define SECTION_STATE	_T("state")
#define SECTION_AUTOKEY	_T("autokey")
#define SECTION_MSGBOARD _T("messageboard")
#define SECTION_MENU	_T("menu")
#define SECTION_ROM     _T("rom")
#define SECTION_COMM    _T("comm")

void init_config()
{
	// initial settings
	memset(&config, 0, sizeof(config_t));

	config.version1 = FILE_VERSION;
	config.version2 = CONFIG_VERSION;

	config.window_mode = 0;
	config.stretch_screen = false;
	config.cutout_screen = false;

	config.sound_frequency = 6;	// 48KHz
	config.sound_latency = 1;	// 100msec

	config.cpu_power = 1;

#ifdef USE_DIPSWITCH
	config.dipswitch = DIPSWITCH_DEFAULT;
#endif
#if defined(USE_MONITOR_TYPE) || defined(USE_SCREEN_ROTATE)
	config.monitor_type = 0;
#endif
#ifdef _BML3MK5
	config.sync_irq = true;
#endif
#ifdef USE_SCANLINE
	config.scan_line = false;
#ifdef _BML3MK5
	config.afterimage = 0;
	config.keepimage = 0;
	config.use_opengl = 0;
	config.disptmg_skew = 0;
	config.curdisp_skew = 0;

	config.volume = 50;
	config.fdd_type = 0;
	config.io_port = 0x44;

	config.show_dlg = 1;

	config.fps_no = -1;

	config.disp_device_no = 0;
	config.screen_width = 0;
	config.screen_height = 0;

	_tcscpy(config.msgboard_fontname, _T("ipagp.ttf"));
	_tcscpy(config.menu_fontname, _T("ipagp.ttf"));

	config.realmode_datarec = false;

	config.screen_video_size = 0;

	config.beep_volume = 50;
	config.psg6_volume = 50;
	config.psg9_volume = 50;
	config.relay_volume = 70;
	config.cmt_volume = 30;
	config.mute = false;
	config.psg6_mute = false;
	config.psg9_mute = false;
	config.relay_mute = false;
	config.cmt_mute = false;

	for(int i=0; i<MAX_PRINTER; i++) {
		_tcscpy(config.printer_server_host[i], _T("localhost"));
		config.printer_server_port[i] = 10200;
		config.printer_direct[i] = false;
	}

	for(int i=0; i<MAX_COMM; i++) {
		config.comm_dipswitch[i] = 1;
		_tcscpy(config.comm_server_host[i], _T("localhost"));
		config.comm_server_port[i] = 6890 + i;
		config.comm_server[i] = false;
		config.comm_connect[i] = false;
	}
#endif
#endif
#ifdef _HC80
	config.device_type = 2;	// Nonintelligent ram disk
#endif
#ifdef _PC8801MA
	config.boot_mode = 2;	// V2 mode, 4MHz
	config.cpu_clock_low = true;
#endif
}

void load_config()
{
	// initial settings
	init_config();

	//
	if (load_ini_file() != true) {
		load_cfg_file();
	}
}

void save_config()
{
	//
	save_ini_file();
	save_cfg_file();
}

void release_config()
{
	delete ini;
	ini = NULL;
}

static bool load_cfg_file()
{
	// get config path
	_TCHAR *app_path, config_path[_MAX_PATH];

	app_path = emu->application_path();
	_stprintf(config_path, _T("%s%s.cfg"), app_path, _T(CONFIG_NAME));

	// load config
	FILEIO* fio = new FILEIO();
	if(fio->Fopen(config_path, FILEIO_READ_BINARY)) {
		fio->Fread((void *)&config, sizeof(config), 1);
		fio->Fclose();
		emu->out_log(LOG_INFO,_T("%s.cfg was loaded."),_T(CONFIG_NAME));
//	} else {
//		emu->out_log(LOG_WARN,_T("%s.cfg couldn't be loaded."),_T(CONFIG_NAME));
	}
	delete fio;

	// check config version
	if(config.version1 != FILE_VERSION) {
		emu->out_log(LOG_WARN,_T("%s.cfg was invalid version."),_T(CONFIG_NAME));
		init_config();
	}
	else if (config.version2 < CONFIG_VERSION) {
		emu->out_log(LOG_WARN,_T("%s.cfg was old version."),_T(CONFIG_NAME));
#ifdef _BML3MK5
		if (config.fdd_type == 1) {
			config.io_port = (config.io_port & ~3) | 2;
		} else if (config.fdd_type == 2) {
			config.io_port = (config.io_port & ~3) | 1;
		}
#endif
		config.version2 = CONFIG_VERSION;
	}
	return true;
}

/// @attention config file no longer use.
static void save_cfg_file()
{
#if 0	// config file no longer use, so don't save this.
	// get config path
	_TCHAR *app_path, config_path[_MAX_PATH];

	app_path = emu->application_path();
	_stprintf(config_path, _T("%s%s.cfg"), app_path, _T(CONFIG_NAME));

	// save config
	FILEIO* fio = new FILEIO();
	if(fio->Fopen(config_path, FILEIO_WRITE_BINARY)) {
		fio->Fwrite((void *)&config, sizeof(config), 1);
		fio->Fclose();
		emu->out_log(LOG_INFO,_T("%s.cfg was saved."),_T(CONFIG_NAME));
	} else {
		emu->out_log(LOG_WARN,_T("%s.cfg couldn't be saved."),_T(CONFIG_NAME));
	}
	delete fio;
#endif
}

static bool load_ini_file()
{
	_TCHAR *app_path, config_path[_MAX_PATH];
	_TCHAR section[100];
	_TCHAR key[100];
	long valuel;
	bool valueb;

	if (ini != NULL) {
		return true;
	}

	ini = new CSimpleIni();

	// load ini file
	app_path = emu->application_path();

	_stprintf(config_path, _T("%s%s.ini"), app_path, _T(CONFIG_NAME));
	if (ini->LoadFile(config_path) == SI_OK) {
		emu->out_log(LOG_INFO,_T("%s.ini was loaded."),_T(CONFIG_NAME));
	} else {
		emu->out_log(LOG_WARN,_T("%s.ini couldn't be loaded."),_T(CONFIG_NAME));
		return false;
	}

	// check config file version
	if (_tcscmp(ini->GetValue(SECTION_MAIN, _T("Name"), _T("")), _T(CLASS_NAME)) != 0) {
		emu->out_log(LOG_WARN,_T("%s.ini was invalid name."),_T(CONFIG_NAME));
		return false;
	}

	config.version1 = ini->GetLongValue(SECTION_MAIN, _T("Version1"));
	if(config.version1 != FILE_VERSION) {
		emu->out_log(LOG_WARN,_T("%s.ini was invalid version."),_T(CONFIG_NAME));
		init_config();
		return false;
	}
	config.version2 = ini->GetLongValue(SECTION_MAIN, _T("Version2"));
	if(config.version2 != CONFIG_VERSION) {
		emu->out_log(LOG_INFO,_T("%s.ini was old version."),_T(CONFIG_NAME));
	}

	valuel = ini->GetLongValue(SECTION_CONTROL, _T("CpuPower"), config.cpu_power);
	if (0 <= valuel && valuel <= 5) {
		config.cpu_power = (uint8)valuel;
	}

#ifdef USE_DIPSWITCH
	config.dipswitch = (uint8)(ini->GetLongValue(SECTION_CONTROL, _T("DipSwitch"), config.dipswitch) & 0xff);
#endif

#ifdef USE_FD1
	_tcsncpy(config.initial_disk_path, ini->GetValue(SECTION_FDD, _T("Path"), config.initial_disk_path), _MAX_PATH-1);

#if defined(USE_FD8) || defined(USE_FD7)
	for (int i=0; i<8; i++) { 
#elif defined(USE_FD6) || defined(USE_FD5)
	for (int i=0; i<6; i++) { 
#else
	for (int i=0; i<4; i++) { 
#endif
		for (int j=0; j<8; j++) { 
			if (config.version2 < 8) {
				// old version is start base 1
				_stprintf(section, _T("%s%d"), SECTION_FDD, i+1);
			} else {
				_stprintf(section, _T("%s%d"), SECTION_FDD, i);
			}
			_stprintf(key, _T("File%d"), j+1);
			_tcsncpy(config.recent_disk_path[i][j], ini->GetValue(section, key, config.recent_disk_path[i][j]), _MAX_PATH-1);
		}
	}
#endif
#ifdef USE_DATAREC
	_tcsncpy(config.initial_datarec_path, ini->GetValue(SECTION_TAPE, _T("Path"), config.initial_datarec_path), _MAX_PATH-1);
	for (int j=0; j<8; j++) { 
		_stprintf(key, _T("File%d"), j+1);
		_tcsncpy(config.recent_datarec_path[j], ini->GetValue(SECTION_TAPE, key, config.recent_datarec_path[j]), _MAX_PATH-1);
	}
	config.realmode_datarec = ini->GetBoolValue(SECTION_TAPE, _T("RealMode"), config.realmode_datarec);
#endif

	config.window_mode = ini->GetLongValue(SECTION_SCREEN, _T("WindowMode"), config.window_mode);
#ifdef _BML3MK5
	config.disp_device_no = ini->GetLongValue(SECTION_SCREEN, _T("DisplayDevice"), config.disp_device_no);
	config.screen_width = ini->GetLongValue(SECTION_SCREEN, _T("ScreenWidth"), config.screen_width);
	config.screen_height = ini->GetLongValue(SECTION_SCREEN, _T("ScreenHeight"), config.screen_height);
#endif
	config.stretch_screen = ini->GetBoolValue(SECTION_SCREEN, _T("StretchScreen"), config.stretch_screen);
	config.cutout_screen = ini->GetBoolValue(SECTION_SCREEN, _T("CutoutScreen"), config.cutout_screen);

	config.sound_frequency = ini->GetLongValue(SECTION_SOUND,  _T("FrequencyNo"), config.sound_frequency);
	config.sound_latency = ini->GetLongValue(SECTION_SOUND,  _T("LatencyNo"), config.sound_latency);

#if defined(USE_MONITOR_TYPE) || defined(USE_SCREEN_ROTATE)
	config.monitor_type = ini->GetLongValue(SECTION_SCREEN,  _T("MonitorType"), config.monitor_type);
#endif
#ifdef USE_SCANLINE
	config.scan_line = ini->GetBoolValue(SECTION_SCREEN, _T("ScanLine"), config.scan_line);
#endif
#ifdef _BML3MK5
	config.sync_irq = ini->GetBoolValue(SECTION_CONTROL, _T("SyncIRQ"), config.sync_irq);

	valuel = ini->GetLongValue(SECTION_SCREEN, _T("AfterImage"), config.afterimage);
	if (0 <= valuel && valuel <= 2) {
		config.afterimage = (uint8)valuel;
	}
	valuel = ini->GetLongValue(SECTION_SCREEN, _T("KeepImage"), config.keepimage);
	if (0 <= valuel && valuel <= 2) {
		config.keepimage = (uint8)valuel;
	}
	valuel = ini->GetLongValue(SECTION_SCREEN, _T("UseOpenGL"), config.use_opengl);
	if (0 <= valuel && valuel <= 2) {
		config.use_opengl = (uint8)valuel;
	}
	ini->Delete(SECTION_SCREEN, _T("UseDirect3D"));	// not use in this app
	valuel = ini->GetLongValue(SECTION_SCREEN, _T("DisptmgSkew"), config.disptmg_skew);
	if (0 <= valuel && valuel <= 2) {
		config.disptmg_skew = (uint8)valuel;
	}
	valuel = ini->GetLongValue(SECTION_SCREEN, _T("CurdispSkew"), config.curdisp_skew);
	if (0 <= valuel && valuel <= 2) {
		config.curdisp_skew = (uint8)valuel;
	}
	valuel = ini->GetLongValue(SECTION_SCREEN, _T("VideoSize"), config.screen_video_size);
	if (0 <= valuel && valuel <= 1) {
		config.screen_video_size = (uint8)valuel;
	}

	valuel = ini->GetLongValue(SECTION_SOUND,  _T("Volume"), config.volume);
	if (0 <= valuel && valuel <= 100) {
		config.volume = (int)valuel;
	}
	valuel = ini->GetLongValue(SECTION_CONTROL, _T("FddType"), config.fdd_type);
	if (0 <= valuel && valuel <= 2) {
		config.fdd_type = (int)valuel;
	}
	config.io_port = ini->GetLongValue(SECTION_CONTROL, _T("IoPort"), config.io_port) & 0x7f;

	valueb = ini->GetBoolValue(SECTION_CONTROL, _T("ShowLED"), (config.show_dlg & 1) ? true : false);
	config.show_dlg = valueb ? (config.show_dlg | 1) : (config.show_dlg & ~1);
	valueb = ini->GetBoolValue(SECTION_CONTROL, _T("ShowMessage"), (config.show_dlg & 2) ? true : false);
	config.show_dlg = valueb ? (config.show_dlg | 2) : (config.show_dlg & ~2);
	valueb = ini->GetBoolValue(SECTION_CONTROL, _T("UseJoystick"), (config.show_dlg & 4) ? true : false);
	config.show_dlg = valueb ? (config.show_dlg | 4) : (config.show_dlg & ~4);

	valuel = ini->GetLongValue(SECTION_CONTROL, _T("FpsNo"), config.fps_no);
	if (-1 <= valuel && valuel <= 6) {
		config.fps_no = (int)valuel;
	}

	_tcsncpy(config.initial_printer_path, ini->GetValue(SECTION_PRINTER, _T("Path"), config.initial_printer_path), _MAX_PATH-1);
	_tcsncpy(config.initial_state_path, ini->GetValue(SECTION_STATE, _T("Path"), config.initial_state_path), _MAX_PATH-1);
#ifdef USE_AUTO_KEY
	_tcsncpy(config.initial_autokey_path, ini->GetValue(SECTION_AUTOKEY, _T("Path"), config.initial_autokey_path), _MAX_PATH-1);
#endif
	_tcsncpy(config.msgboard_fontname, ini->GetValue(SECTION_MSGBOARD, _T("FontName"), config.msgboard_fontname), _MAX_PATH-1);
	_tcsncpy(config.menu_fontname, ini->GetValue(SECTION_MENU, _T("FontName"), config.menu_fontname), _MAX_PATH-1);

	_tcsncpy(config.rom_path, ini->GetValue(SECTION_ROM, _T("Path"), config.rom_path), _MAX_PATH-1);

	valuel = ini->GetLongValue(SECTION_SOUND, _T("BeepVolume"), config.beep_volume);
	if (0 <= valuel && valuel <= 100) {
		config.beep_volume = (int)valuel;
	}
	valuel = ini->GetLongValue(SECTION_SOUND, _T("6VoicePsgVolume"), config.psg6_volume);
	if (0 <= valuel && valuel <= 100) {
		config.psg6_volume = (int)valuel;
	}
	valuel = ini->GetLongValue(SECTION_SOUND, _T("9VoicePsgVolume"), config.psg9_volume);
	if (0 <= valuel && valuel <= 100) {
		config.psg9_volume = (int)valuel;
	}
	valuel = ini->GetLongValue(SECTION_SOUND, _T("RelayVolume"), config.relay_volume);
	if (0 <= valuel && valuel <= 100) {
		config.relay_volume = (int)valuel;
	}
	valuel = ini->GetLongValue(SECTION_SOUND, _T("CmtVolume"), config.cmt_volume);
	if (0 <= valuel && valuel <= 100) {
		config.cmt_volume = (int)valuel;
	}
	config.mute = ini->GetBoolValue(SECTION_SOUND, _T("Mute"), config.mute);
	config.beep_mute = ini->GetBoolValue(SECTION_SOUND, _T("BeepMute"), config.beep_mute);
	config.psg6_mute = ini->GetBoolValue(SECTION_SOUND, _T("6VoicePsgMute"), config.psg6_mute);
	config.psg9_mute = ini->GetBoolValue(SECTION_SOUND, _T("9VoicePsgMute"), config.psg9_mute);
	config.relay_mute = ini->GetBoolValue(SECTION_SOUND, _T("RelayMute"), config.relay_mute);
	config.cmt_mute = ini->GetBoolValue(SECTION_SOUND, _T("CmtMute"), config.cmt_mute);

	for (int i=0; i<MAX_PRINTER; i++) { 
		_stprintf(section, _T("%s%d"), SECTION_PRINTER, i);
		_tcsncpy(config.printer_server_host[i], ini->GetValue(section, _T("ServerHost"), config.printer_server_host[i]), _MAX_PATH-1);
		valuel = ini->GetLongValue(section, _T("ServerPort"), config.printer_server_port[i]);
		if (0 <= valuel && valuel <= 65535) {
			config.printer_server_port[i] = (int)valuel;
		}
	}

	for (int i=0; i<MAX_COMM; i++) { 
		_stprintf(section, _T("%s%d"), SECTION_COMM, i);
		valuel = ini->GetLongValue(section, _T("DipSwitch"), config.comm_dipswitch[i]);
		if (1 <= valuel && valuel <= 4) {
			config.comm_dipswitch[i] = (int)valuel;
		}
		_tcsncpy(config.comm_server_host[i], ini->GetValue(section, _T("ServerHost"), config.comm_server_host[i]), _MAX_PATH-1);
		valuel = ini->GetLongValue(section, _T("ServerPort"), config.comm_server_port[i]);
		if (0 <= valuel && valuel <= 65535) {
			config.comm_server_port[i] = (int)valuel;
		}
	}

	if (config.fdd_type == 1) {
		// use 3inch fdd
		config.io_port = (config.io_port & ~3) | 2;
	} else if (config.fdd_type == 2) {
		// use 5inch fdd
		config.io_port = (config.io_port & ~3) | 1;
	}

	// convert from old version data
	if (config.version2 < 8) {
		config.cpu_power++;
	}
#endif
	config.version2 = CONFIG_VERSION;

	return true;
}

static void save_ini_file()
{
	_TCHAR *app_path, config_path[_MAX_PATH];
	_TCHAR comment[100];
	_TCHAR section[100];
	_TCHAR key[100];

	if (ini == NULL) {
		return;
	}

	// section
	_stprintf(comment, _T("; %s config file"), _T(DEVICE_NAME));

	// values
	ini->SetValue(SECTION_MAIN, _T("Name"), _T(CLASS_NAME), comment);
	ini->SetLongValue(SECTION_MAIN, _T("Version1"), config.version1, NULL, true);
	ini->SetLongValue(SECTION_MAIN, _T("Version2"), config.version2);

	ini->SetLongValue(SECTION_CONTROL, _T("CpuPower"), config.cpu_power);

#ifdef USE_DIPSWITCH
	ini->SetLongValue(SECTION_CONTROL, _T("DipSwitch"), config.dipswitch, NULL, true);
#endif


#ifdef USE_FD1
	ini->SetValue(SECTION_FDD, _T("Path"), config.initial_disk_path);

#if defined(USE_FD8) || defined(USE_FD7)
	for (int i=0; i<8; i++) { 
#elif defined(USE_FD6) || defined(USE_FD5)
	for (int i=0; i<6; i++) { 
#else
	for (int i=0; i<4; i++) { 
#endif
		for (int j=0; j<8; j++) { 
			_stprintf(section, _T("%s%d"), SECTION_FDD, i);
			_stprintf(key, _T("File%d"), j+1);
			if (_tcslen(config.recent_disk_path[i][j]) > 0) {
				ini->SetValue(section, key, config.recent_disk_path[i][j]);
			}
		}
	}
#endif
#ifdef USE_DATAREC
	ini->SetValue(SECTION_TAPE, _T("Path"), config.initial_datarec_path);
	for (int j=0; j<8; j++) { 
		_stprintf(key, _T("File%d"), j+1);
		if (_tcslen(config.recent_datarec_path[j]) > 0) {
			ini->SetValue(SECTION_TAPE, key, config.recent_datarec_path[j]);
		}
	}
	ini->SetBoolValue(SECTION_TAPE, _T("RealMode"), config.realmode_datarec);
#endif

	ini->SetLongValue(SECTION_SCREEN, _T("WindowMode"), config.window_mode);
#ifdef _BML3MK5
	ini->SetLongValue(SECTION_SCREEN, _T("DisplayDevice"), config.disp_device_no);
	ini->SetLongValue(SECTION_SCREEN, _T("ScreenWidth"), config.screen_width);
	ini->SetLongValue(SECTION_SCREEN, _T("ScreenHeight"), config.screen_height);
#endif
	ini->SetBoolValue(SECTION_SCREEN, _T("StretchScreen"), config.stretch_screen);
	ini->SetBoolValue(SECTION_SCREEN, _T("CutoutScreen"), config.cutout_screen);

	ini->SetLongValue(SECTION_SOUND,  _T("FrequencyNo"), config.sound_frequency);
	ini->SetLongValue(SECTION_SOUND,  _T("LatencyNo"), config.sound_latency);

#if defined(USE_MONITOR_TYPE) || defined(USE_SCREEN_ROTATE)
	ini->SetLongValue(SECTION_SCREEN,  _T("MonitorType"), config.monitor_type);
#endif
#ifdef USE_SCANLINE
	ini->SetBoolValue(SECTION_SCREEN, _T("ScanLine"), config.scan_line);
#endif
#ifdef _BML3MK5
	ini->SetBoolValue(SECTION_CONTROL, _T("SyncIRQ"), config.sync_irq);

	ini->SetLongValue(SECTION_SCREEN, _T("AfterImage"), config.afterimage);
	ini->SetLongValue(SECTION_SCREEN, _T("KeepImage"), config.keepimage);
	ini->SetLongValue(SECTION_SCREEN, _T("UseOpenGL"), config.use_opengl);
	ini->SetLongValue(SECTION_SCREEN, _T("DisptmgSkew"), config.disptmg_skew);
	ini->SetLongValue(SECTION_SCREEN, _T("CurdispSkew"), config.curdisp_skew);
	ini->SetLongValue(SECTION_SCREEN, _T("VideoSize"), config.screen_video_size);

	ini->SetLongValue(SECTION_SOUND,  _T("Volume"), config.volume);
	ini->SetLongValue(SECTION_CONTROL, _T("FddType"), config.fdd_type);
	ini->SetLongValue(SECTION_CONTROL, _T("IoPort"), config.io_port, NULL, true);

	ini->SetBoolValue(SECTION_CONTROL, _T("ShowLED"), (config.show_dlg & 1) ? true : false);
	ini->SetBoolValue(SECTION_CONTROL, _T("ShowMessage"), (config.show_dlg & 2) ? true : false);
	ini->SetBoolValue(SECTION_CONTROL, _T("UseJoystick"), (config.show_dlg & 4) ? true : false);

	ini->SetLongValue(SECTION_CONTROL, _T("FpsNo"), config.fps_no);

	ini->SetValue(SECTION_PRINTER, _T("Path"), config.initial_printer_path);
	ini->SetValue(SECTION_STATE, _T("Path"), config.initial_state_path);
#ifdef USE_AUTO_KEY
	ini->SetValue(SECTION_AUTOKEY, _T("Path"), config.initial_autokey_path);
#endif
	ini->SetValue(SECTION_MSGBOARD, _T("FontName"), config.msgboard_fontname);
	ini->SetValue(SECTION_MENU, _T("FontName"), config.menu_fontname);
	ini->SetValue(SECTION_ROM, _T("Path"), config.rom_path);
#endif

	ini->SetLongValue(SECTION_SOUND, _T("BeepVolume"), config.beep_volume);
	ini->SetLongValue(SECTION_SOUND, _T("6VoicePsgVolume"), config.psg6_volume);
	ini->SetLongValue(SECTION_SOUND, _T("9VoicePsgVolume"), config.psg9_volume);
	ini->SetLongValue(SECTION_SOUND, _T("RelayVolume"), config.relay_volume);
	ini->SetLongValue(SECTION_SOUND, _T("CmtVolume"), config.cmt_volume);
	ini->SetBoolValue(SECTION_SOUND, _T("Mute"), config.mute);
	ini->SetBoolValue(SECTION_SOUND, _T("BeepMute"), config.beep_mute);
	ini->SetBoolValue(SECTION_SOUND, _T("6VoicePsgMute"), config.psg6_mute);
	ini->SetBoolValue(SECTION_SOUND, _T("9VoicePsgMute"), config.psg9_mute);
	ini->SetBoolValue(SECTION_SOUND, _T("RelayMute"), config.relay_mute);
	ini->SetBoolValue(SECTION_SOUND, _T("CmtMute"), config.cmt_mute);

	for(int i=0; i<MAX_PRINTER; i++) {
		_stprintf(section, _T("%s%d"), SECTION_PRINTER, i);
		ini->SetValue(section, _T("ServerHost"), config.printer_server_host[i]);
		ini->SetLongValue(section, _T("ServerPort"), config.printer_server_port[i]);
	}

	for(int i=0; i<MAX_COMM; i++) {
		_stprintf(section, _T("%s%d"), SECTION_COMM, i);
		ini->SetLongValue(section, _T("DipSwitch"), config.comm_dipswitch[i]);
		ini->SetValue(section, _T("ServerHost"), config.comm_server_host[i]);
		ini->SetLongValue(section, _T("ServerPort"), config.comm_server_port[i]);
	}

	// save ini file
	app_path = emu->application_path();

	_stprintf(config_path, _T("%s%s.ini"), app_path, _T(CONFIG_NAME));
	if (ini->SaveFile(config_path) == SI_OK) {
		emu->out_log(LOG_INFO,_T("%s.ini was saved."),_T(CONFIG_NAME));
	} else {
		emu->out_log(LOG_WARN,_T("%s.ini couldn't be saved."),_T(CONFIG_NAME));
	}
}

