/** @file emu.h

	Skelton for retropc emulator
	SDL edition

	@author Sasaji
	@date   2012.02.21

	@brief [ sdl emulation i/f ]

	@note
	This code is based on the Common Source Code Project.
	Original Author : Takeda.Toshiya
*/

#ifndef _EMU_H_
#define _EMU_H_


// for log
// #define _LOG_CONSOLE
#define _LOG_FILE

#define LOG_INFO	0
#define LOG_WARN	1
#define LOG_ERROR	2
#define LOG_DEBUG	3

// for debug
//#define _DEBUG_LOG
#ifdef _DEBUG_LOG
	// output debug log to console
//	#define _DEBUG_CONSOLE
	// output debug log to file
//	#define _DEBUG_FILE

	// output cpu debug log
//	#define _CPU_DEBUG_LOG
	// output fdc debug log
//	#define _FDC_DEBUG_LOG
	// output i/o debug log
//	#define _IO_DEBUG_LOG
#endif

#include <SDL.h>
#include "common.h"
#include "config.h"
#include "vm/vm.h"
#include "ledbox.h"
#include "msgboard.h"

#ifndef SCREEN_WIDTH_ASPECT
#define SCREEN_WIDTH_ASPECT SCREEN_WIDTH
#endif
#ifndef SCREEN_HEIGHT_ASPECT
#define SCREEN_HEIGHT_ASPECT SCREEN_HEIGHT
#endif
#ifndef WINDOW_WIDTH
#define WINDOW_WIDTH SCREEN_WIDTH_ASPECT
#endif
#ifndef WINDOW_HEIGHT
#define WINDOW_HEIGHT SCREEN_HEIGHT_ASPECT
#endif

#ifdef USE_OPENGL
#include <SDL_opengl.h>
#endif

#ifdef USE_SOCKET
#include <SDL_net.h>
#endif

#ifdef USE_MEDIA
#define MEDIA_MAX 64
#endif
#ifdef USE_SOCKET
#define SOCKET_MAX 6
#define SOCKET_BUFFER_MAX 0x100000
#define SDL_USEREVENT_SOCKET (SDL_USEREVENT + 1)
#define SDL_USEREVENT_SOCKET_CONNECTED 1
#define SDL_USEREVENT_SOCKET_DISCONNECTED 2
#define SDL_USEREVENT_SOCKET_WRITEABLE 3
#define SDL_USEREVENT_SOCKET_READABLE 4
#define SDL_USEREVENT_SOCKET_ACCEPT 5
#endif

#define KEY_STATUS_SIZE 512
#define KEY_MOD_ALT_KEY 0x0001

class FIFO;
class FILEIO;
class AG_CONFIG_DLG;
class AG_KEYBIND_DLG;
class AG_GUI_BASE;

/**
	@brief emulation i/f
*/
class EMU
{
protected:
	VM* vm;
	AG_GUI_BASE *gui;
private:
	// ----------------------------------------
	// input
	// ----------------------------------------
	/// @name input private methods
	//@{
	void initialize_input();
	void release_input();
	void update_input();

	void initialize_joystick();
	void release_joystick();
	//@}
	/// @name input private members
	//@{
	uint8 key_status[KEY_STATUS_SIZE];	// windows key code mapping
#ifdef USE_SHIFT_NUMPAD_KEY
	uint8 key_converted[256];
	bool key_shift_pressed, key_shift_released;
#endif
	bool lost_focus;
	int  key_mod;

	uint32 joy_status[2];	// joystick #1, #2 (b0 = up, b1 = down, b2 = left, b3 = right, b4-b31 = trigger #1-#28
	int joy_num;
	int joy_xmin[2], joy_xmax[2];
	int joy_ymin[2], joy_ymax[2];
	bool joy_enabled[2];
	SDL_Joystick *joy[2];
	bool use_joystick;

	int mouse_status[3];	// x, y, button (b0 = left, b1 = right)
	bool mouse_enabled;

#ifdef USE_AUTO_KEY
	FIFO* autokey_buffer;
	int autokey_phase, autokey_shift;
	uint16 autokey_code;
#ifdef _BML3MK5
	int autokey_graph;
#endif
#endif
	//@}
public:
	/// @name input public methods
	//@{
	void reset_joystick();
	void update_joystick();
	//@}

private:
	// ----------------------------------------
	// screen
	// ----------------------------------------
	/// @name screen private methods
	//@{
	void EMU_SCREEN();
	void initialize_screen();
	void release_screen();
	inline void copy_surface(SDL_Surface *src, SDL_Rect *reSrc, SDL_Surface *dst, SDL_Rect *reDst);
	inline void zoomin_surface(SDL_Surface *src, SDL_Rect *reSrc, SDL_Surface *dst, SDL_Rect *reDst);

#ifdef USE_OPENGL
	void initialize_opengl();
	void release_opengl();

	void set_pixel_format();
	void set_opengl_attr();
	void set_opengl_poly(int width, int height);
#endif
	//@}
	/// @name screen private static methods
	//@{
	static int qsort_by_w_h_n(const void *a, const void *b);
	static int qsort_by_n_w_h(const void *a, const void *b);
	//@}
	/// @name screen private members
	//@{
	// screen settings
	SDL_PixelFormat pixel_format;
	SDL_Surface *screen;
	SDL_Surface *srcsurface;
	SDL_Surface *mixsurface;
	Uint32 screen_flags;

	/// vm draw area
	int screen_width, screen_height;
	int screen_width_aspect, screen_height_aspect;
	/// vm draw area (not used)
	int window_width, window_height;
	/// minimum window client size
	int min_window_width, min_window_height;
	/// display size (window client size)
	int display_width, display_height;
	/// desktop size on the monitor
	int desktop_width, desktop_height;
	bool screen_size_changed;

	SDL_Rect rec_video_size[2];

	/// vm draw area
	/// if use screen rotation, swap width and height.
	int source_width, source_height;
	int source_width_aspect, source_height_aspect;
	int stretched_width, stretched_height;
	int stretch_pow_x, stretch_pow_y;
	int screen_dest_x, screen_dest_y;
	int display_stretched_width, display_stretched_height;
	int screen_stretched_dest_x, screen_stretched_dest_y;

	// update flags
	bool first_draw_screen;
	bool first_invalidate;
	bool self_invalidate;
	bool skip_frame;
	bool skip_frame_pre;

	// screen buffer
	scrntype* lpBmp;

	SDL_Rect reMix;
	SDL_Rect reSuf;

	// screen mode
#define SCREEN_MODE_MAX 24
	int screen_mode_count;
	struct screen_mode_st {
		int disp_no;
		int width;
		int height;
	} screen_mode[SCREEN_MODE_MAX];

#define WINDOW_MODE_MAX 8
	int window_mode_count;
	struct window_mode_st {
		int power;
		int width;
		int height;
	} window_mode[WINDOW_MODE_MAX];
	int window_mode_power;

	bool now_fullscreen;
	int  prev_window_mode;

	// record video
	bool now_rec_vid;
	int rec_frames, rec_fps;

#ifdef USE_OPENGL
	GLuint mix_texture_name;
	SDL_Rect reTex;
	int use_opengl;
#endif
	//@}

	// ----------------------------------------
	// sound
	// ----------------------------------------
	/// @name sound private methods
	//@{
	void initialize_sound(int rate, int samples, int latency);
	void initialize_sound(int rate, int samples);
	void initialize_sound();
	void release_sound();
	void start_sound();
	void stop_sound();
//	void update_sound(int* extra_frames);
	//@}
	/// @name sound private static methods
	//@{
	// callback
	static void update_sound(void *userdata, Uint8 *stream, int len);
	//@}
	/// @name sound private members
	//@{
	SDL_AudioSpec audio_desired, audio_obtained;

	int sound_rate, sound_samples;
	int sound_latency;
	bool sound_ok, now_mute;

	// direct sound
	uint32 sound_prev_time;

	// record sound
	typedef struct {
		uint32 dwRIFF;
		uint32 dwFileSize;
		uint32 dwWAVE;
		uint32 dwfmt_;
		uint32 dwFormatSize;
		uint16 wFormatTag;
		uint16 wChannels;
		uint32 dwSamplesPerSec;
		uint32 dwAvgBytesPerSec;
		uint16 wBlockAlign;
		uint16 wBitsPerSample;
		uint32 dwdata;
		uint32 dwDataLength;
	} wavheader_t;
	FILEIO* rec;
	int rec_bytes;
	bool now_rec_snd;
	//@}

	// ----------------------------------------
	// floppy disk
	// ----------------------------------------
#ifdef USE_FD1
	/// @name floppy disk private methods
	//@{
	void initialize_disk_insert();
	void update_disk_insert();
	//@}
	/// @name floppy disk private members
	//@{
	typedef struct {
		_TCHAR path[_MAX_PATH];
		int offset;
		int wait_count;
		uint32 flags;
	} disk_insert_t;
	disk_insert_t disk_insert[8];
	//@}
#endif

	// ----------------------------------------
	// media
	// ----------------------------------------
#ifdef USE_MEDIA
	/// @name media private methods
	//@{
	void initialize_media();
	void release_media();
	//@}
	/// @name media private members
	//@{
	_TCHAR media_path[MEDIA_MAX][_MAX_PATH];
	int media_cnt;
	bool media_playing;
	//@}
#endif

	// ----------------------------------------
	// timer
	// ----------------------------------------
	/// @name timer private methods
	//@{
	void update_timer();
	//@}
	/// @name timer private members
	//@{
//	SYSTEMTIME sTime;
	struct tm sTime;
	//@}

	// ----------------------------------------
	// socket
	// ----------------------------------------
#ifdef USE_SOCKET
	/// @name socket private methods
	//@{
	void initialize_socket();
	void release_socket();
	void update_socket();
	//@}
	/// @name socket private members
	//@{
//	SOCKET soc[SOCKET_MAX];
	bool is_tcp[SOCKET_MAX];
//	struct sockaddr_in udpaddr[SOCKET_MAX];
	int socket_delay[SOCKET_MAX];
	char *recv_buffer[SOCKET_MAX];
	int recv_r_ptr[SOCKET_MAX], recv_w_ptr[SOCKET_MAX];
	int socket_nums;

	SDLNet_SocketSet socket_set;
	TCPsocket tcp_socket[SOCKET_MAX];
	UDPsocket udp_socket[SOCKET_MAX];
	UDPpacket udp_packet;
	bool is_server[SOCKET_MAX];
	//@}
#endif

	// ----------------------------------------
	// misc
	// ----------------------------------------
	/// @name misc private methods
	//@{
	void open_logfile();
	void close_logfile();
#ifdef _BML3MK5
	int update_led();
#endif
	//@}
	/// @name misc private members
	//@{
#ifdef _LOG_FILE
	FILE* fplog;
#endif
	_TCHAR app_path[_MAX_PATH];

	bool vm_pause,vm_syspause,vm_usrpause;

	LedBox *ledbox;
	MsgBoard *msgboard;
	AG_CONFIG_DLG *configbox;
	//@}

public:
	AG_KEYBIND_DLG *keybindbox;

//
//
//
public:
	// ----------------------------------------
	// initialize
	// ----------------------------------------
	/// @name initialize
	//@{
	EMU(const char *path);
	~EMU();
	void initialize();
	void release();
	//@}

	// ----------------------------------------
	// misc
	// ----------------------------------------
	/// @name misc
	//@{
	void application_path(_TCHAR *path);
	_TCHAR *application_path(void);
	bool get_pause(int idx);
	void set_pause(int idx, bool val);
	bool *get_pause_ptr(void);
	void set_gui(AG_GUI_BASE *new_gui);
	//@}

	// ----------------------------------------
	// for windows
	// ----------------------------------------
	/// @name drive virtual machine
	//@{
	int frame_interval();
	int run(int split_num);
	void update();
	bool now_skip();
	//@}
	// user interface
	/// @name control menu for ui
	//@{
	void update_config();
	void reset();
#ifdef USE_SPECIAL_RESET
	void special_reset();
	void warm_reset();
#endif
	void change_dipswitch(int num);
#ifdef USE_POWER_OFF
	void notify_power_off();
#endif
	void change_cpu_power(int num);
#ifdef _BML3MK5
	void change_sync_irq();
#endif
#ifdef USE_CART
	void open_cart(_TCHAR* file_path);
	void close_cart();
#endif
#ifdef USE_AUTO_KEY
	void start_auto_key();
	void start_auto_key(_TCHAR *path);
	void stop_auto_key();
	bool now_auto_key() {
		return (autokey_phase != 0);
	}
#endif
#ifdef _BML3MK5
	void save_state(_TCHAR* filename);
	void load_state(_TCHAR* filename);
#endif
	//@}
#ifdef USE_FD1
	/// @name floppy disk menu for ui
	//@{
	bool open_disk(int drv, _TCHAR* file_path, int offset, uint32 flags);
	bool open_disk(int drv, _TCHAR* file_path, int offset);
	void close_disk(int drv);
	int  change_disk(int drv);
	int  get_disk_side(int drv);
	//@}
#endif
#ifdef USE_QUICKDISK
	void open_quickdisk(_TCHAR* file_path);
	void close_quickdisk();
#endif
#ifdef USE_DATAREC
	/// @name tape menu for ui
	//@{
	bool play_datarec(_TCHAR* file_path);
	bool rec_datarec(_TCHAR* file_path);
	void close_datarec();
	void rewind_datarec();
	void fast_forward_datarec();
	void realmode_datarec();
	//@}
#endif
#ifdef USE_DATAREC_BUTTON
	void push_play();
	void push_stop();
#endif
#ifdef USE_MEDIA
	void open_media(_TCHAR* file_path);
	void close_media();
#endif
#ifdef USE_BINARY_FILE1
	void load_binary(int drv, _TCHAR* file_path);
	void save_binary(int drv, _TCHAR* file_path);
#endif
	/// @name screen menu for ui
	//@{
	void change_screen_mode(int mode);
	void change_stretch_screen();
	void change_cutout_screen();
	int get_window_mode_count() {
		return window_mode_count;
	}
	int get_screen_mode_count() {
		return screen_mode_count;
	}
	int is_fullscreen() {
		return now_fullscreen;
	}
	void get_window_mode_str(int num, char *str);
	void get_screen_mode_str(int num, char *str);
	void get_screen_record_size_str(int num, char *str);
	void capture_screen();
	void start_rec_video(int fps, bool show_dialog);
	void stop_rec_video();
	void restart_rec_video();
	void resize_rec_video(int new_size);
	bool now_rec_video() {
		return now_rec_vid;
	}
#ifdef USE_SCANLINE
	void change_screen_scanline();
#endif
#ifdef _BML3MK5
#ifdef USE_OPENGL
	void change_screen_use_opengl(int num);
#endif
	void change_screen_afterimage(int num);
	void change_screen_keepimage(int num);
#endif
	//@}
	/// @name sound menu for ui
	//@{
	void start_rec_sound();
	void stop_rec_sound();
	void restart_rec_sound();
	bool now_rec_sound() {
		return now_rec_snd;
	}
	bool *get_now_rec_sound_ptr() {
		return &now_rec_snd;
	}
	void set_volume(int volume);
	//@}
#ifdef USE_PRINTER
	/// @name printer menu for ui
	//@{
	void save_printer(int drv, _TCHAR* filename);
	void clear_printer(int drv);
	int  get_printer_buffer_size(int drv);
	uint8* get_printer_buffer(int drv);
//	void send_printer_data(uint8 *datas, int len);
	void enable_printer_direct(int drv);
	bool print_printer(int drv);
	//@}
#endif
#ifdef _BML3MK5
	/// @name comm menu for ui
	//@{
	void enable_comm_server(int drv);
	void enable_comm_connect(int drv);
	//@}
#endif
#ifdef _BML3MK5
	/// @name options menu for ui
	//@{
	void show_led();
	void move_led();
	bool is_shown_ledbox(int id);
	void show_message_board();
	bool is_shown_message_board();
	void toggle_use_joypad();
	bool is_enable_joypad();

	void save_keybind();
	//@}
#endif

	//
	/// @name input device procedures for host machine
	//@{
//	void key_down(int code, bool repeat);
	int  key_down(int sym, uint16 mod);
//	void key_up(int code);
	void key_up(int sym, uint16 mod);
	void key_lost_focus() {
		lost_focus = true;
	}
	void translate_sdl_keysym(SDL_Event *e, int *new_type, int *new_sym);
	void translate_sdl_keysym(Uint8 e_type, Uint8 e_scancode, SDLKey e_sym, SDLMod e_mod, Uint16 e_unicode, int *new_type, int *new_sym);
	bool translate_key_down(int sym, uint16 mod);

#ifdef USE_BUTTON
	void press_button(int num);
#endif

	void enable_mouse();
	void disenable_mouse();
	void toggle_mouse();
	bool get_mouse_enabled() {
		return mouse_enabled;
	}

#ifdef USE_AUTO_KEY
	void parse_auto_key(char *buf, int size, int &prev_code, int &prev_kana);
#endif
//	void enable_joystick();
	//@}
	/// @name screen device procedures for host machine
	//@{
	int get_window_width(int mode);
	int get_window_height(int mode);
	bool create_screen(int width, int height, Uint32 flags);
	void set_display_size(int width, int height, int power, bool window_mode);
	void change_screen_size(int sw, int sh, int swa, int sha, int ww, int wh);
	void draw_screen();
	void mix_screen();
	void update_screen();
	void skip_screen() {
		skip_frame = true;
	}

	void enum_window_mode(int max_width, int max_height);
	void enum_screen_mode(SDL_PixelFormat *format, Uint32 flags);
	void find_screen_mode();
	void set_window(int mode, int cur_width, int cur_height);
	void set_desktop_size(int width, int height) {
		desktop_width = width;
		desktop_height = height;
	}

#ifdef USE_BITMAP
	void reload_bitmap() {
		first_invalidate = true;
	}
#endif
	bool create_offlinesurface();

#ifdef USE_OPENGL
	int  get_use_opengl() {
		return use_opengl;
	}
	void set_use_opengl(int val) {
		use_opengl = val;
	}
#endif
	//@}
	/// @name sound device procedures for host machine
	//@{
	void mute_sound(bool);
	//@}
#ifdef USE_SOCKET
	/// @name socket device procedures for host machine
	//@{
//	int get_socket(int ch);
	void *get_socket(int ch);
	void socket_connected(int ch);
	void socket_disconnected(int ch);
	void socket_writeable(int ch);
	void socket_readable(int ch);
	void socket_accept(int ch);
	void send_data(int ch);
	void recv_data(int ch);
	//@}
#endif

	// ----------------------------------------
	// for virtual machine
	// ----------------------------------------

	// power off
	void power_off() {
//		PostMessage(main_window_handle, WM_CLOSE, 0, 0L);
	}

	/// @name input device for vm
	//@{
	uint8* key_buffer() {
		return key_status;
	}
	uint32* joy_buffer() {
		return joy_status;
	}
	int* mouse_buffer() {
		return mouse_status;
	}
	int* get_key_mod_ptr() {
		return &key_mod;	
	}
	void recognized_key(uint16 key_code);
	//@}
	/// @name screen for vm
	//@{
	scrntype* screen_buffer(int y);
	int screen_buffer_offset();
	void get_rgbformat(uint32 *r_mask, uint32 *g_mask, uint32 *b_mask, uint8 *r_shift, uint8 *g_shift, uint8 *b_shift);
	void get_rgbcolor(scrntype pixel, uint8 *r, uint8 *g, uint8 *b);
	scrntype map_rgbcolor(uint8 r, uint8 g, uint8 b);
	bool* get_skip_frame() {
		return &skip_frame;
	}
	//@}
	/// @name sound for vm
	//@{
	void record_rec_sound(uint8 *buffer, int samples);
	void record_rec_sound(uint16 *buffer, int samples);
	void record_rec_sound(int32 *buffer, int samples);
	//@}
	/// @name timer for vm
	//@{
	void get_timer(int time[]);
	//@}
#ifdef USE_MEDIA
	/// @name media for vm
	//@{
	int media_count();
	void play_media(int trk);
	void stop_media();
	//@}
#endif
#ifdef USE_SOCKET
	/// @name socket for vm
	//@{
	bool init_socket_tcp(int ch, bool server = false);
	bool init_socket_udp(int ch, bool server = false);
	bool connect_socket(int ch, uint32 ipaddr, int port, bool server = false);
	bool connect_socket(int ch, const _TCHAR *hostname, int port, bool server = false);
	bool connect_socket(int ch, IPaddress *ip, bool server = false);
	bool get_ipaddr(const _TCHAR *hostname, int port, IPaddress *ip);
	void disconnect_socket(int ch);
	int  get_socket_channel();
	bool listen_socket(int ch);
	void send_data_tcp(int ch);
	void send_data_udp(int ch, uint32 ipaddr, int port);
	void send_data_udp(int ch, const _TCHAR *hostname, int port);
	//@}
#endif
#ifdef _BML3MK5
//	void send_comm_data(uint8 *data, size_t size);
//	void recv_comm_data(uint8 *data, size_t size);
	/// for get_parami method
	enum enumParamId {
		ConfigBoxFddType,
		ConfigBoxWavReverse,
		ConfigBoxWavHalf,
		ConfigBoxWavCorrect,
		ConfigBoxWavCorrectType,
		ConfigBoxWavSampleRate,
		ConfigBoxWavSampleBits,
		ConfigBoxIOPort,		
	};
	/// @name misc for vm
	//@{
	int get_parami(enumParamId id);
	//@}
#endif
	/// @name logging methods
	//@{
	void out_log(int level, const _TCHAR* format, ...);
	void out_logv(int level, const _TCHAR* format, va_list ap);
	/// debug log
	void out_debug(const _TCHAR* format, ...);
	//
	void out_msg(const _TCHAR* msg, bool set, int sec);
	void out_msg(const _TCHAR* msg, bool set = true);
	void out_info(const _TCHAR* msg, bool set, int sec);
	void out_info(const _TCHAR* msg, bool set = true);
	void out_infof(const _TCHAR* format, ...);
	//@}
};

#endif
