/** @file sdl_input.cpp

	Skelton for retropc emulator
	SDL edition

	@author Sasaji
	@date   2012.02.21

	@brief [ sdl input ]

	@note
	This code is based on win32_input.cpp of the Common Source Code Project.
	Author : Takeda.Toshiya
*/

#include "emu.h"
#include "vm/vm.h"
#include "fifo.h"
#include "fileio.h"

#define KEY_KEEP_FRAMES 3

void EMU::initialize_input()
{
	out_debug(_T("EMU::initialize_input"));

	// initialize status
	memset(key_status, 0, sizeof(key_status));
	memset(joy_status, 0, sizeof(joy_status));
	memset(mouse_status, 0, sizeof(mouse_status));

	for(int i=0; i<2; i++) {
		joy[i] = NULL;
		joy_enabled[i] = false;
	}
	use_joystick = (config.show_dlg & 4) ? true : false;
	// initialize joysticks
	if (SDL_InitSubSystem(SDL_INIT_JOYSTICK)) {
		out_log(LOG_WARN, _T("SDL_InitSubSystem(SDL_INIT_JOYSTICK): %s."), SDL_GetError());
	}
	initialize_joystick();

	// mouse emulation is disenabled
	mouse_enabled = false;

#ifdef USE_AUTO_KEY
	// initialize autokey
	autokey_buffer = new FIFO(65536);
	autokey_buffer->clear();
	autokey_phase = autokey_shift = 0;
#ifdef _BML3MK5
	autokey_graph = 0;
#endif
#define AUTO_KEY_KANA	SDLK_TAB
#define AUTO_KEY_GRAPH	SDLK_F8
#endif
	lost_focus = false;
	key_mod = 0;
}

void EMU::initialize_joystick()
{
#ifdef USE_JOYSTICK
	for(int i = 0; i < 2; i++) {
		joy_xmin[i] = -16384;
		joy_xmax[i] = 16384;
		joy_ymin[i] = -16384;
		joy_ymax[i] = 16384;
	}
	SDL_JoystickEventState(SDL_IGNORE);
	reset_joystick();
#endif
}

void EMU::reset_joystick()
{
#ifdef USE_JOYSTICK
	if (!use_joystick) return;

	joy_num = SDL_NumJoysticks();
	for(int i = 0; i < joy_num && i < 2; i++) {
		if (!SDL_JoystickOpened(i)) {
			if ((joy[i] = SDL_JoystickOpen(i)) != NULL) {
				joy_enabled[i] = true;
			}
		}
	}
#endif
}

void EMU::release_input()
{
	// release mouse
	if(mouse_enabled) {
		disenable_mouse();
	}
	// release joystick
	release_joystick();

#ifdef USE_AUTO_KEY
	// release autokey buffer
	if(autokey_buffer) {
		autokey_buffer->release();
		delete autokey_buffer;
	}
#endif
}

void EMU::release_joystick()
{
#ifdef USE_JOYSTICK
	// release joystick
	for(int i = 0; i < 2; i++) {
		if (joy[i] != NULL) {
			SDL_JoystickClose(joy[i]);
		}
		joy[i] = NULL;
		joy_enabled[i] = false;
	}
#endif
}

void EMU::update_input()
{
//	emu->out_debug("EMU::update_input");

	// release keys
#ifdef USE_AUTO_KEY
	if(lost_focus && autokey_phase == 0) {
#else
	if(lost_focus) {
#endif
		// we lost key focus so release all pressed keys
		for(int i = 0; i < KEY_STATUS_SIZE; i++) {
			if(key_status[i] & 0x80) {
				key_status[i] &= 0x7f;
#ifdef NOTIFY_KEY_DOWN
				if(!key_status[i]) {
					vm->key_up(i);
				}
#endif
			}
		}
	}
	else {
		for(int i = 0; i < KEY_STATUS_SIZE; i++) {
			if(key_status[i] & 0x7f) {
				key_status[i] = (key_status[i] & 0x80) | ((key_status[i] & 0x7f) - 1);
#ifdef NOTIFY_KEY_DOWN
				if(!key_status[i]) {
					vm->key_up(i);
				}
#endif
			}
		}
	}
	lost_focus = false;

#ifdef USE_JOYSTICK
	update_joystick();
#endif
#ifdef USE_KEY_TO_JOY
	// emulate joystick #1 with keyboard
	if(key_status[0x26]) joy_status[0] |= 0x01;	// up
	if(key_status[0x28]) joy_status[0] |= 0x02;	// down
	if(key_status[0x25]) joy_status[0] |= 0x04;	// left
	if(key_status[0x27]) joy_status[0] |= 0x08;	// right
#endif

	// update mouse status
	memset(mouse_status, 0, sizeof(mouse_status));
	if(mouse_enabled) {
		// get current status
		Point pt;
		Uint8 mstat = SDL_GetMouseState(&(pt.x),&(pt.y));

		mouse_status[0]  = pt.x - display_width / 2;
		mouse_status[1]  = pt.y - display_height / 2;
		mouse_status[2] =  (mstat & SDL_BUTTON(SDL_BUTTON_LEFT) ? 1 : 0);
		mouse_status[2] |= (mstat & SDL_BUTTON(SDL_BUTTON_RIGHT) ? 2 : 0);
		mouse_status[2] |= (mstat & SDL_BUTTON(SDL_BUTTON_MIDDLE) ? 4 : 0);
		// move mouse cursor to the center of window
		if(!(mouse_status[0] == 0 && mouse_status[1] == 0)) {
			pt.x = display_width / 2;
			pt.y = display_height / 2;
			SDL_WarpMouse(pt.x, pt.y);
		}
	}
#ifdef USE_LIGHTPEN
	if(true) {
		// get current status
		Point pt;
		Uint8 mstat = SDL_GetMouseState(&(pt.x),&(pt.y));
		pt.x = screen_width * (pt.x - screen_dest_x) / stretched_width;
		pt.y = screen_height * (pt.y - screen_dest_y) / stretched_height;

		mouse_status[0] = pt.x;
		mouse_status[1] = pt.y;
		mouse_status[2] = (mstat & SDL_BUTTON(SDL_BUTTON_LEFT) ? 1 : 0) | (mstat & SDL_BUTTON(SDL_BUTTON_RIGHT) ? 2 : 0);
	}
#endif

#ifdef USE_AUTO_KEY
	// auto key
	switch(autokey_phase) {
	case 1:
		if(autokey_buffer && !autokey_buffer->empty()) {
			// update shift key status
			int shift = autokey_buffer->read_not_remove(0) & 0x400;
			if(shift && !autokey_shift) {
				key_down(SDLK_RSHIFT, KMOD_SHIFT);
			}
			autokey_shift = shift;
#ifdef _BML3MK5
			int graph = autokey_buffer->read_not_remove(0) & 0x800;
			if(graph && !autokey_graph) {
				key_down(AUTO_KEY_GRAPH, 0);
			}
			autokey_graph = graph;
#endif
		}
		autokey_phase++;
		break;
	case 3:
		if(autokey_buffer && !autokey_buffer->empty()) {
//			key_down(autokey_buffer->read_not_remove(0) & 0x3ff, 0);
			autokey_code = autokey_buffer->read_not_remove(0) & 0x3ff;
			autokey_code = key_down(autokey_code, 0);
		}
		out_debug(_T("autokeydown: phase:%d auto_key:%x"),autokey_phase,autokey_code);
		autokey_phase++;
		break;
	case 4:
		if(!autokey_buffer || autokey_buffer->empty()) {
			autokey_phase++;
		}
		// wait response from vm
		break;
	case USE_AUTO_KEY:
		if(autokey_shift) {
			key_up(SDLK_RSHIFT, 0);
			autokey_shift = 0;
		}
#ifdef _BML3MK5
		if(autokey_graph) {
			key_up(AUTO_KEY_GRAPH, 0);
			autokey_graph = 0;
		}
#endif
		if(autokey_buffer && !autokey_buffer->empty()) {
			key_up(autokey_buffer->read_not_remove(0) & 0x3ff, 0);
		}
		out_debug(_T("autokeyup: phase:%d auto_key:%x"),autokey_phase,autokey_code);
		autokey_code = 0;
		autokey_phase++;
		break;
	case USE_AUTO_KEY + 1:
		if(!autokey_buffer || autokey_buffer->empty()) {
			autokey_phase++;
		}
		// wait response from vm
		break;
	case USE_AUTO_KEY_RELEASE:
		if(autokey_buffer && !autokey_buffer->empty()) {
			// wait enough while vm analyzes one line
			if(autokey_buffer->read() == 0xd) {
				autokey_phase++;
				break;
			}
		}
		// through
	case 30:
		if(autokey_buffer && !autokey_buffer->empty()) {
			autokey_phase = 1;
		}
		else {
			stop_auto_key();
		}
		break;
	default:
		if(autokey_phase) {
			autokey_phase++;
		}
		break;
	}
#endif
}

void EMU::recognized_key(uint16 key_code)
{
	if(autokey_phase == 0) return;
	if(autokey_phase == 4 && autokey_code) {
		out_debug(_T("recognized_key: phase:%d auto_key:%x vm_key:%x"),autokey_phase,autokey_code,key_code);
		if (autokey_code == key_code) autokey_phase++;
	}
	if(autokey_phase == USE_AUTO_KEY + 1 && key_code == 0) {
		out_debug(_T("recognized_key: phase:%d auto_key:%x vm_key:%x"),autokey_phase,autokey_code,key_code);
		autokey_phase++;
	}
}

void EMU::update_joystick()
{
#ifdef USE_JOYSTICK
	if (!use_joystick) return;

	// update joystick status
	SDL_JoystickUpdate();
	memset(joy_status, 0, sizeof(joy_status));
	for(int i = 0; i < 2; i++) {
		if(joy_enabled[i] == true) {
			if (SDL_JoystickOpened(i)) {
				int x = SDL_JoystickGetAxis(joy[i], 0);
				int y = SDL_JoystickGetAxis(joy[i], 1);
				if(y < joy_ymin[i]) joy_status[i] |= 0x01;
				if(y > joy_ymax[i]) joy_status[i] |= 0x02;
				if(x < joy_xmin[i]) joy_status[i] |= 0x04;
				if(x > joy_xmax[i]) joy_status[i] |= 0x08;
				int nbuttons = SDL_JoystickNumButtons(joy[i]);
				for(int n=0; n<nbuttons && n<28; n++) {
					if (SDL_JoystickGetButton(joy[i], n)) {
						joy_status[i] |= (0x10 << n);
					}
				}
			} else {
				joy_enabled[i] = false;
			}
		}
	}
#endif
}

int EMU::key_down(int code, uint16 mod)
{
	if (!translate_key_down(code, mod)) {
		key_status[code] = 0x80;
	}
#ifdef NOTIFY_KEY_DOWN
	vm->key_down(code);
#endif
	return code;
}

void EMU::key_up(int code, uint16 mod)
{
	if(key_status[code]) {
		key_status[code] &= 0x7f;
#ifdef NOTIFY_KEY_DOWN
		if(!key_status[code]) {
			vm->key_up(code);
		}
#endif
	}
}

#ifdef USE_BUTTON
void EMU::press_button(int num)
{
	int code = buttons[num].code;

	if(code) {
		key_down(code, false);
		key_status[code] = KEY_KEEP_FRAMES * FRAME_SPLIT_NUM;
	}
	else {
		// code=0: reset virtual machine
		vm->reset();
	}
}
#endif

void EMU::enable_mouse()
{
	// enable mouse emulation
	if(!mouse_enabled) {
		// hide mouse cursor
		SDL_ShowCursor(SDL_DISABLE);
		// move mouse cursor to the center of window
		Point pt;
		pt.x = display_width / 2;
		pt.y = display_height / 2;
		SDL_WarpMouse(pt.x, pt.y);
	}
	mouse_enabled = true;
}

void EMU::disenable_mouse()
{
	// disenable mouse emulation
	if(mouse_enabled) {
		SDL_ShowCursor(SDL_ENABLE);
	}
	mouse_enabled = false;
}

void EMU::toggle_mouse()
{
	// toggle mouse enable / disenable
	if(mouse_enabled) {
		disenable_mouse();
	}
	else {
		enable_mouse();
	}
}

#ifdef USE_AUTO_KEY
static const int autokey_table[256] = {
#ifdef _BML3MK5
	// 0x400: shift
	// 0x800: graph
	// 0x1000: katakana
	0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x00d,0x000,0x000,0x00d,0x000,0x000,
	0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,
	0x020,0x431,0x432,0x433,0x434,0x435,0x436,0x437,0x438,0x439,0x43a,0x43b,0x02c,0x02d,0x02e,0x02f,
	0x030,0x031,0x032,0x033,0x034,0x035,0x036,0x037,0x038,0x039,0x03a,0x03b,0x42c,0x42d,0x42e,0x42f,
	0x040,0x061,0x062,0x063,0x064,0x065,0x066,0x067,0x068,0x069,0x06a,0x06b,0x06c,0x06d,0x06e,0x06f,
	0x070,0x071,0x072,0x073,0x074,0x075,0x076,0x077,0x078,0x079,0x07a,0x05b,0x05c,0x05d,0x05e,0x05f,
	0x000,0x461,0x462,0x463,0x464,0x465,0x466,0x467,0x468,0x469,0x46a,0x46b,0x46c,0x46d,0x46e,0x46f,
	0x470,0x471,0x472,0x473,0x474,0x475,0x476,0x477,0x478,0x479,0x47a,0x000,0x000,0x000,0x000,0x000,
	// hiragana1
	0x87a,0x878,0x863,0x876,0x862,0x86e,0x86d,0x82c,0x82e,0x86c,0x86f,0x839,0x838,0x869,0x86b,0x905,
	0x902,0x908,0x906,0x904,0x83b,0x90d,0x90e,0x82f,0x907,0x909,0x901,0x903,0x872,0x874,0x866,0x867,
	// katakana
	0x0831,0x142e,0x145b,0x145d,0x142c,0x142f,0x1430,0x1433,0x1465,0x1434,0x1435,0x1436,0x1437,0x1438,0x1439,0x147a,
	0x105c,0x1033,0x1065,0x1034,0x1035,0x1036,0x1074,0x1067,0x1068,0x103a,0x1062,0x1078,0x1064,0x1072,0x1070,0x1063,
	0x1071,0x1061,0x107a,0x1077,0x1073,0x1075,0x1069,0x1031,0x102c,0x106b,0x1066,0x1076,0x1032,0x105e,0x102d,0x106a,
	0x106e,0x105d,0x102f,0x106d,0x1037,0x1038,0x1039,0x106f,0x106c,0x102e,0x103b,0x105f,0x1030,0x1079,0x1040,0x105b,
	// hiragana2
	0x82d,0x840,0x83a,0x85f,0x879,0x875,0x868,0x86a,0x871,0x861,0x877,0x873,0x865,0x864,0x90b,0x90c,
	0x900,0x85c,0x832,0x833,0x834,0x835,0x836,0x837,0x90a,0x830,0x870,0x85e,0x85b,0x2079,0x85d,0x000,
#else
	// 0x100: shift
	// 0x200: kana
	// 0x400: alphabet
	// 0x800: ALPHABET
	0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x00d,0x000,0x000,0x00d,0x000,0x000,
	0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,
	0x020,0x131,0x132,0x133,0x134,0x135,0x136,0x137,0x138,0x139,0x1ba,0x1bb,0x0bc,0x0bd,0x0be,0x0bf,
	0x030,0x031,0x032,0x033,0x034,0x035,0x036,0x037,0x038,0x039,0x0ba,0x0bb,0x1bc,0x1bd,0x1be,0x1bf,
	0x0c0,0x441,0x442,0x443,0x444,0x445,0x446,0x447,0x448,0x449,0x44a,0x44b,0x44c,0x44d,0x44e,0x44f,
	0x450,0x451,0x452,0x453,0x454,0x455,0x456,0x457,0x458,0x459,0x45a,0x0db,0x0dc,0x0dd,0x0de,0x1e2,
	0x1c0,0x841,0x842,0x843,0x844,0x845,0x846,0x847,0x848,0x849,0x84a,0x84b,0x84c,0x84d,0x84e,0x84f,
	0x850,0x851,0x852,0x853,0x854,0x855,0x856,0x857,0x858,0x859,0x85a,0x1db,0x1dc,0x1dd,0x1de,0x000,
	0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,
	0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,
	// kana -->
	0x000,0x3be,0x3db,0x3dd,0x3bc,0x3bf,0x330,0x333,0x345,0x334,0x335,0x336,0x337,0x338,0x339,0x35a,
	0x2dc,0x233,0x245,0x234,0x235,0x236,0x254,0x247,0x248,0x2ba,0x242,0x258,0x244,0x252,0x250,0x243,
	0x251,0x241,0x25a,0x257,0x253,0x255,0x249,0x231,0x2bc,0x24b,0x246,0x256,0x232,0x2de,0x2bd,0x24a,
	0x24e,0x2dd,0x2bf,0x24d,0x237,0x238,0x239,0x24f,0x24c,0x2be,0x2bb,0x2e2,0x230,0x259,0x2c0,0x2db,
	// <--- kana
	0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,
	0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000
#endif
};

void EMU::start_auto_key()
{
#if 0
	stop_auto_key();

	if(OpenClipboard(NULL)) {
		HANDLE hClip = GetClipboardData(CF_TEXT);
		if(hClip) {
			autokey_buffer->clear();
			int prev_code = 0;
			char* buf = (char*)GlobalLock(hClip);
			int size = strlen(buf), prev_kana = 0;
			parse_auto_key(buf, size, prev_code, prev_kana);
			if(prev_kana) {
				autokey_buffer->write(AUTO_KEY_KANA);
#ifdef _BML3MK5
				if (prev_kana == 0x1000) autokey_buffer->write(AUTO_KEY_KANA);
#endif
			}
			GlobalUnlock(hClip);

			autokey_phase = 1;
			autokey_shift = 0;
#ifdef _BML3MK5
			autokey_graph = 0;
#endif
		}
		CloseClipboard();
	}
#endif
}

void EMU::start_auto_key(_TCHAR *file_path)
{
	FILEIO *fp;
	char buf[1025];

	stop_auto_key();

	fp = new FILEIO();
	int prev_kana = 0;
	if (fp->Fopen(file_path, FILEIO_READ_BINARY)) {
		autokey_buffer->clear();
		int prev_code = 0;
		int size = 0;
		while((size = (int)fp->Fread(buf, sizeof(char), 1024)) > 0) {
			parse_auto_key(buf, size, prev_code, prev_kana);
		}
		fp->Fclose();

		if(prev_kana) {
			autokey_buffer->write(AUTO_KEY_KANA);
#ifdef _BML3MK5
			if (prev_kana == 0x1000) autokey_buffer->write(AUTO_KEY_KANA);
#endif
		}

		autokey_phase = 1;
		autokey_shift = 0;
#ifdef _BML3MK5
		autokey_graph = 0;
#endif
	}
	delete fp;
}

void EMU::parse_auto_key(char *buf, int size, int &prev_code, int &prev_kana)
{
	for(int i = 0; i < size; i++) {
		int code = buf[i] & 0xff;
#ifdef _BML3MK5
		if(prev_code == 0xd && code == 0xa) {
			prev_code = code;
			continue;	// cr-lf
		}
#else
		if((0x81 <= code && code <= 0x9f) || 0xe0 <= code) {
			i++;	// kanji ?
			continue;
		}
		else if(code == 0xa) {
			continue;	// cr-lf
		}
#endif
		if((code = autokey_table[code]) != 0) {
#ifdef _BML3MK5
			int kana = code & 0x3000;
			if(kana - prev_kana == 0x1000) {
				autokey_buffer->write(AUTO_KEY_KANA);
			} else if(kana - prev_kana == 0x2000) {
				autokey_buffer->write(AUTO_KEY_KANA);
				autokey_buffer->write(AUTO_KEY_KANA);
			}
			if(prev_kana - kana == 0x2000) {
				autokey_buffer->write(AUTO_KEY_KANA);
			} else if(prev_kana - kana == 0x1000) {
				autokey_buffer->write(AUTO_KEY_KANA);
				autokey_buffer->write(AUTO_KEY_KANA);
			}
#else
			int kana = code & 0x200;
			if(prev_kana != kana) {
				autokey_buffer->write(AUTO_KEY_KANA);
			}
#endif
			prev_kana = kana;
			if(code & 0x400) {
				autokey_buffer->write((code & 0x3ff) | 0x400);
			}
#ifdef _BML3MK5
			else if (code & 0x800) {
				autokey_buffer->write((code & 0x3ff) | 0x800);
			}
#endif
			else {
				autokey_buffer->write(code & 0x3ff);
			}
		}
		prev_code =code;
	}
}

void EMU::stop_auto_key()
{
	if(autokey_shift) {
		key_up(SDLK_LSHIFT, 0);
	}
	autokey_phase = autokey_shift = 0;
#ifdef _BML3MK5
	if(autokey_graph) {
		key_up(SDLK_WORLD_1, 0);
	}
	autokey_graph = 0;

	memset(key_status, 0, sizeof(key_status));
#endif
}
#endif
