/** @file sdl_input_keysym.cpp

	Skelton for retropc emulator
	SDL edition

	@author Sasaji
	@date   2012.03.08

	@brief [ sdl input keysym ]
*/

#include "emu.h"
#include "vm/vm.h"
#include <SDL.h>

#define KEY_KEEP_FRAMES 3

//
void EMU::translate_sdl_keysym(SDL_Event *e, int *new_type, int *new_sym)
{
	translate_sdl_keysym(e->key.type, e->key.keysym.scancode, e->key.keysym.sym, e->key.keysym.mod, e->key.keysym.unicode, new_type, new_sym);
}

/// Translate key codes from US keylayout to JP keylayout.
void EMU::translate_sdl_keysym(Uint8 e_type, Uint8 e_scancode, SDLKey e_sym, SDLMod e_mod, Uint16 e_unicode, int *new_type, int *new_sym)
{
	int n_type = e_type;
	int n_sym = e_sym;

#if defined(_WIN32)
	//
	// for windows
	//
	// L[CAEgUSzȂ̂ŋJISzɂB
	switch(e_sym) {
	case SDLK_EQUALS:
		n_sym = SDLK_CARET;
		break;
	case SDLK_QUOTE:
		n_sym = SDLK_COLON;
		break;
	case SDLK_LEFTBRACKET:
		n_sym = SDLK_AT;
		break;
	case SDLK_BACKSLASH:
		switch(e_scancode) {
		case 0x2b:
			n_sym = SDLK_RIGHTBRACKET;
			break;
		}
		break;
	case SDLK_RIGHTBRACKET:
		n_sym = SDLK_LEFTBRACKET;
		break;
	case SDLK_CAPSLOCK:
		if (e_type == SDL_KEYUP && (e_mod & KMOD_CAPS) == 0) {
			n_type = SDL_KEYDOWN;
		}
		break;
	default:
		switch(e_scancode) {
		case 0x70:	// katakana
			n_sym = SDLK_WORLD_2;
			break;
		case 0x79:	// henkan
			n_sym = SDLK_WORLD_0;
			break;
		case 0x73:	// underscore
			n_sym = SDLK_UNDERSCORE;
			break;
		case 0x7b:	// muhenkan
			n_sym = SDLK_WORLD_1;
			break;
		}
		break;
	}
	out_debug(_T("[win] translate_sdl_keysym: type:%02x -> %02x sym:%02x -> %02x")
		,e_type, n_type
		,e_sym, n_sym);

#elif defined(linux)
	//
	// for linux
	//
	switch(e_sym) {
	default:
		switch(e_scancode) {
		case 0x42:	// caps lock
			n_sym = SDLK_CAPSLOCK;
			break;
		case 0x64:	// henkan
			n_sym = SDLK_WORLD_0;
			break;
		case 0x65:	// katakana / hiragana
			n_sym = SDLK_WORLD_2;
			break;
		case 0x61:	// underscore
			n_sym = SDLK_UNDERSCORE;
			break;
		case 0x66:	// muhenkan
			n_sym = SDLK_WORLD_1;
			break;
		}
		break;
	}
	out_debug("[linux] translate_sdl_keysym: type:%02x -> %02x sym:%02x -> %02x"
		,e_type, n_type
		,e_sym, n_sym);

#elif defined(__APPLE__) && defined(__MACH__)
	//
	// for MacOS X
	//
	// L[CAEgUSzȂ̂ŋJISzɂB
	switch(e_sym) {
	case SDLK_EQUALS:
		n_sym = SDLK_CARET;
		break;
	case SDLK_QUOTE:
		n_sym = SDLK_COLON;
		break;
	case SDLK_LEFTBRACKET:
		n_sym = SDLK_AT;
		break;
	case SDLK_BACKSLASH:
		n_sym = SDLK_RIGHTBRACKET;
		break;
	case SDLK_RIGHTBRACKET:
		n_sym = SDLK_LEFTBRACKET;
		break;
	case SDLK_CAPSLOCK:
		if (e_type == SDL_KEYUP && (e_mod & KMOD_CAPS) == 0) {
			n_type = SDL_KEYDOWN;
		}
		n_sym = e_sym;
		break;
	default:
		switch(e_scancode) {
		case 0x5d:	// yen
			n_sym = SDLK_BACKSLASH;
			break;
		case 0x5e:	// underscore
			n_sym = SDLK_UNDERSCORE;
			break;
		case 0x6e:	// right win
			n_sym = SDLK_MENU;
			break;
		}
		break;
	}
	out_debug("[mac] translate_sdl_keysym: type:%02x -> %02x sym:%02x -> %02x"
		,e_type, n_type
		,e_sym, n_sym);

#elif defined(__FreeBSD__)
	//
	// for FreeBSD
	//
	switch(e_sym) {
	default:
		switch(e_scancode) {
		case 0x31:	// kanji
			n_sym = SDLK_BACKQUOTE;
			break;
		case 0x42:	// caps lock
			n_sym = SDLK_CAPSLOCK;
			break;
		case 0x81:	// henkan
			n_sym = SDLK_WORLD_0;
			break;
		case 0x83:	// muhenkan
			n_sym = SDLK_WORLD_1;
			break;
		case 0xd0:	// katakana / hiragana
			n_sym = SDLK_WORLD_2;
			break;
		case 0xd3:	// underscore
			n_sym = SDLK_UNDERSCORE;
			break;
		}
		break;
	}
	out_debug("[unix] translate_sdl_keysym: type:%02x -> %02x sym:%02x -> %02x"
		,e_type, n_type
		,e_sym, n_sym);

#endif
	if (new_type != NULL) *new_type = n_type;
	if (new_sym  != NULL) *new_sym  = n_sym;
}

bool EMU::translate_key_down(int code, uint16 mod)
{
#if defined(_WIN32) || (defined(__APPLE__) && defined(__MACH__))
	if(code == SDLK_CAPSLOCK) {
		key_status[code] = KEY_KEEP_FRAMES * FRAME_SPLIT_NUM;
		return true;
	}
//	else if(code == SDLK_WORLD_0) {
//		key_status[code] = KEY_KEEP_FRAMES * FRAME_SPLIT_NUM;
//	}
#endif

	return false;
}