/** @file sdl_main.cpp

	Skelton for retropc emulator
	SDL edition

	@author Sasaji
	@date   2012.2.15

	@brief [ sdl_main ]

	@note
	This code is based on winmain.cpp of the Common Source Code Project.
	Author : Takeda.Toshiya
*/

#include <agar/core.h>
#include "common.h"
#include "config.h"
#include "emu.h"
#include "gui/ag_gui.h"
#include "gui/ag_file_dlg.h"

#define MIN_SKIP_FRAMES 0
#define MAX_SKIP_FRAMES 10

#define MIN_INTERVAL	4

#define CALC_NEXT_INTERVAL(now_ms, interval_ms, delta_ms) { \
	interval_ms = now_ms / FRAMES_PER_SEC; \
	now_ms = now_ms - (interval_ms * FRAMES_PER_SEC) + delta_ms; \
}

// variable
EMU *emu = NULL;
AG_GUI_BASE *gui = NULL;


int fskip_remain = 0;
static const int fps[6] = {60, 30, 20, 15, 12, 10};
static const int fskip[6] = {1, 2, 3, 4, 5, 6};
int rec_fps_no = -1;


bool working = true;
bool gui_process = false;

int *key_mod = NULL;

static char app_path[_MAX_PATH];
static char app_name[_MAX_PATH];

static char tape_file[_MAX_PATH];

int desktop_width = 0;
int desktop_height = 0;

// prototype
static void exit_program(void);

static void main_event_loop();
static void vm_event_loop();

static void vm_event_proc(SDL_Event *);

static int get_options(int, char *[]);

void get_dir_and_basename(const char *, char *, char *);



/*
 *	main
 */
int main(int argc, char* argv[])
{
#if defined(_DEBUG) && defined(_WIN32) && !defined(__MINGW32__) 
	_CrtSetDbgFlag( _CRTDBG_ALLOC_MEM_DF | _CRTDBG_LEAK_CHECK_DF );
#endif

	char buf[_MAX_PATH];

	const SDL_VideoInfo *video = NULL;
//	char icon_file[_MAX_PATH];

	// get and parse options
	get_options(argc, argv);

	// create a instance of emulation core
	emu = new EMU(app_path);

	// load config
	load_config();

	gui = new AG_GUI(emu);
	if (gui->InitCore() == -1) {
		delete gui;
		exit(1);
	}
	emu->set_gui(gui);

	// disable keylock
//	SDL_putenv("SDL_DISABLE_LOCK_KEYS=1");

	// init SDL
	Uint32 init_flags = SDL_INIT_VIDEO | SDL_INIT_AUDIO | SDL_INIT_TIMER;
#ifndef _DEBUG
	init_flags |= SDL_INIT_NOPARACHUTE;
#endif
	if(SDL_Init(init_flags)) {
		emu->out_log(LOG_ERROR, _T("SDL_Init: %s."), SDL_GetError());
		exit(1);
	}
	atexit(exit_program);

	// UNICODE support
	if (SDL_EnableUNICODE(1) == -1) {
		emu->out_log(LOG_WARN, _T("SDL_EnableUNICODE: %s."), SDL_GetError());
	}

#if !(defined(__APPLE__) && defined(__MACH__))
	// window icon
	sprintf(buf, "%s/res/%s.bmp", app_path, CONFIG_NAME);
	SDL_WM_SetIcon(SDL_LoadBMP(buf), NULL);
#endif

	// get current video mode
	if((video = SDL_GetVideoInfo()) == NULL) {
		emu->out_log(LOG_ERROR, _T("SDL_GetVideoInfo: %s."), SDL_GetError());
		exit(1);
	}
#ifdef _DEBUG_LOG
	emu->out_debug(_T("video->hw_available:%d"),video->hw_available);
	emu->out_debug(_T("video->wm_available:%d"),video->wm_available);
	emu->out_debug(_T("video->blit_hw:%d"),video->blit_hw);
	emu->out_debug(_T("video->blit_hw_CC:%d"),video->blit_hw_CC);
	emu->out_debug(_T("video->blit_hw_A:%d"),video->blit_hw_A);
	emu->out_debug(_T("video->blit_sw:%d"),video->blit_sw);
	emu->out_debug(_T("video->blit_sw_CC:%d"),video->blit_sw_CC);
	emu->out_debug(_T("video->blit_sw_A:%d"),video->blit_sw_A);
	emu->out_debug(_T("video->blit_fill:%d"),video->blit_fill);
	emu->out_debug(_T("video->video_mem:%d"),video->video_mem);
	emu->out_debug(_T("video->vfmt->BitsPerPixel:%d"),video->vfmt->BitsPerPixel);
	emu->out_debug(_T("video->vfmt->BytesPerPixel:%d"),video->vfmt->BytesPerPixel);
	emu->out_debug(_T("video->vfmt->Rmask:%x, Gmask:%x, Bmask:%x, Amask:%x"),video->vfmt->Rmask,video->vfmt->Gmask,video->vfmt->Bmask,video->vfmt->Amask);
	emu->out_debug(_T("video->vfmt->Rshift:%d, Gshift:%d, Bshift:%d, Ashift:%d"),video->vfmt->Rshift,video->vfmt->Gshift,video->vfmt->Bshift,video->vfmt->Ashift);
	emu->out_debug(_T("video->vfmt->Rloss:%d, Gloss:%d, Bloss:%d, Aloss:%d"),video->vfmt->Rloss,video->vfmt->Gloss,video->vfmt->Bloss,video->vfmt->Aloss);
	emu->out_debug(_T("video->vfmt->colorkey:%d"),video->vfmt->colorkey);
	emu->out_debug(_T("video->vfmt->alpha:%d"),video->vfmt->alpha);
	emu->out_debug(_T("video->current_w:%d"),video->current_w);
	emu->out_debug(_T("video->current_h:%d"),video->current_h);
#endif
	desktop_width = video->current_w;
	desktop_height = video->current_h;

	Uint32 video_flags = 0;
	if (video->hw_available) {
		video_flags |= SDL_HWSURFACE;
	}
#if (defined(__APPLE__) && defined(__MACH__))
	// for mac
	video_flags |= SDL_ASYNCBLIT;
#endif
#ifdef USE_OPENGL
	emu->set_use_opengl(config.use_opengl);
#endif
	// create window
	emu->set_desktop_size(desktop_width, desktop_height);
	emu->enum_window_mode(desktop_width, desktop_height);
	emu->enum_screen_mode(video->vfmt, video_flags);
	emu->find_screen_mode();
	emu->set_window(config.window_mode, desktop_width, desktop_height);

	if (!emu->create_screen(config.screen_width, config.screen_height, video_flags)) {
		exit(1);
	}
	// menu is created in create_screen method
//	// create gui menu
//	if (gui->CreateMenu() == -1) {
//		exit(1);
//	}
	// create global keys
	if (gui->CreateGlobalKeys() == -1) {
		exit(1);
	}

	// initialize emulation core
	emu->initialize();
	if (!emu->create_offlinesurface()) {
		exit(1);
	}
	key_mod = emu->get_key_mod_ptr();

	// open files
	if (emu) {
		if (tape_file[0] != '\0') {
			gui->filebox->LoadDataRecFile(tape_file);
		}
#ifdef USE_FD1
		// auto open recent file
		for(int i=0; i<MAX_DRIVE; i++) {
			if (config.recent_disk_path[i][0][0] != '\0') {
				gui->filebox->OpenFloppyFile(i, config.recent_disk_path[i][0], 0);
			}
		}
#endif
	}

	// vm start
	if (emu) {
		emu->reset();
	}

	// main event loop
	main_event_loop();

	// release emulation core
	if(emu) {
		emu->release();
	}

	// save config
	save_config();
	release_config();

	return 0;
}

/*
 *	quit
 */
static void exit_program(void)
{
	if (gui) {
		delete gui;
	}
	if (emu) {
		delete emu;
		emu = NULL;
	}
	SDL_Quit();
}

/*
 *	main event loop
 */
static void main_event_loop()
{
	vm_event_loop();
}

/*
 * 	vm event loop (SDL)
 */
static void vm_event_loop()
{
	SDL_Event event;

	int total_frames = 0, draw_frames = 0, skip_frames = 0;

	int ms = 1000;
	int next_interval;
	CALC_NEXT_INTERVAL(ms, next_interval, 1000);

	int split_num = 0;

	uint32 start_time = 0;
	uint32 current_time = 0;
	int    remain_time = 0;
	uint32 next_time = 0;
	uint32 update_fps_time = next_time + 1000;

	int msgcnt = 0;
	char buf[256];

	// play sound
	emu->mute_sound(false);

	// process coming messages at first. 
	while(SDL_PollEvent(&event)) {
		int rc = gui->ProcessEvent(&event);
		if (rc < 0) {
			working = false;
			break;
		} else if (rc > 0) {
			gui_process = false;
			gui->Show(gui_process);
		}
		vm_event_proc(&event);
	}
	gui->ProcessTimeout();

	next_time = SDL_GetTicks();
	// loop
	while(working) {
		split_num = 0;
		// get next period
		CALC_NEXT_INTERVAL(ms, next_interval, 1000);

		start_time = next_time;
//		next_time += emu->now_skip() ? 0 : next_interval;
		next_time += next_interval;
		current_time = SDL_GetTicks();

		// when vm processing is too late...
		if (next_time + 200 <= current_time) {
			// adjust next time for current time
//			next_time = current_time + (emu->now_skip() ? 0 : next_interval);
			next_time = current_time + next_interval;
		}
//		emu->out_log(LOG_DEBUG, "A1: cur:%ld st:%ld nt:%ld ms:%d int:%d", current_time, start_time, next_time, ms, next_interval);

		for(int i=0; i < (FRAME_SPLIT_NUM - 1); i++) {
			// drive machine
			if(emu)	emu->run(split_num);

			msgcnt = 10;
			while (SDL_PollEvent(&event) && msgcnt > 0) {
				int rc = gui->ProcessEvent(&event);
				if (rc < 0) {
					working = false;
					goto FIN;
					break;
				} else if (rc > 0) {
					gui_process = false;
					gui->Show(gui_process);
				}
				vm_event_proc(&event);
				msgcnt--;
			}
			gui->ProcessTimeout();

			split_num++;
		}

		if(emu) {
			// drive machine
			emu->run(split_num);

//			if (split_num + 1 == FRAME_SPLIT_NUM) {
				total_frames++;
				if(emu->now_rec_video()) {
					if (fskip_remain <= 0) {
						// rec pictures constant frames per 1 second
						emu->mix_screen();
						if (gui_process) {
							gui->MixGuiScreen();
						}
						emu->update_screen();
						draw_frames++;
						skip_frames = 0;
						fskip_remain = fskip[rec_fps_no];
					} else {
						skip_frames++;
						emu->skip_screen();
					}
					if (fskip_remain > 0) fskip_remain--;
				}
				else if(config.fps_no >= 0) {
					if (fskip_remain <= 0) {
						// constant frames per 1 second
						emu->mix_screen();
						if (gui_process) {
							gui->MixGuiScreen();
						}
						emu->update_screen();
						draw_frames++;
						skip_frames = 0;
						fskip_remain = fskip[config.fps_no];
					} else {
						skip_frames++;
						emu->skip_screen();
					}
					if (fskip_remain > 0) fskip_remain--;
				}
				else {
					current_time = SDL_GetTicks();
					if(next_time > current_time) {
						// update window if enough time
						emu->mix_screen();
						if (gui_process) {
							gui->MixGuiScreen();
						}
						emu->update_screen();
						draw_frames++;
						skip_frames = 0;
					}
					else if(++skip_frames > MAX_SKIP_FRAMES) {
						// update window at least once per 10 frames
						emu->mix_screen();
						if (gui_process) {
							gui->MixGuiScreen();
						}
						emu->update_screen();
						draw_frames++;
						skip_frames = 0;
						next_time = SDL_GetTicks();
					}
					else {
						emu->skip_screen();
					}
				}
//			}
		}
		// process event at left time in 1 frame if need
		do {
//			current_time = SDL_GetTicks();
//			remain_time = next_time - current_time;
			// check window message
			msgcnt = 100;
			while (SDL_PollEvent(&event) && msgcnt > 0) {
				int rc = gui->ProcessEvent(&event);
				if (rc < 0) {
					working = false;
					goto FIN;
					break;
				} else if (rc > 0) {
					gui_process = false;
					gui->Show(gui_process);
				}
				vm_event_proc(&event);
				msgcnt--;
			}
			gui->ProcessTimeout();

			if (skip_frames == 0) {
				current_time = SDL_GetTicks();
				remain_time = next_time - current_time;
				if (remain_time > MIN_INTERVAL) {
					if (emu) emu->update();
					current_time = SDL_GetTicks();
					remain_time = next_time - current_time;
					if (remain_time > MIN_INTERVAL) {
						SDL_Delay(MIN_INTERVAL);
						current_time = SDL_GetTicks();
						remain_time = next_time - current_time;
					}
				}
			} else {
				remain_time = 0;
			}
		} while (remain_time > 1);

//		split_num = (split_num + 1) & (FRAME_SPLIT_NUM - 1);

//		emu->out_debug("A2: rti:%ld nt:%ld (total:%d fps:%d skip:%d)",SDL_GetTicks(),next_time,total_frames,draw_frames,skip_frames);
//		emu->out_log(LOG_DEBUG,"A2: rti:%ld nt:%ld",SDL_GetTicks(),next_time);

		// calc frame rate
		current_time = SDL_GetTicks();
		if(update_fps_time <= current_time) {
			int ratio = 0;
			if (total_frames) ratio = (int)(100.0 * (double)draw_frames / (double)total_frames + 0.5);
			sprintf(buf, "%s - %d / %d fps (%d %%)", DEVICE_NAME, draw_frames, total_frames, ratio);
			SDL_WM_SetCaption(buf, NULL);

			update_fps_time += 1000;
			if(update_fps_time <= current_time) {
				update_fps_time = current_time + 1000;
			}
			total_frames = draw_frames = 0;
		}
	}

FIN:
	return;
}

/*
 *	vm event procedure (SDL)
 */
static void vm_event_proc(SDL_Event *e)
{
	switch(e->type) {
	case SDL_KEYDOWN:
	case SDL_KEYUP:
		if(emu) {
			int   type = e->key.type;
			uint16 mod = e->key.keysym.mod;
			int    sym = e->key.keysym.sym;
			emu->translate_sdl_keysym(e,&type,&sym);
			switch(type) {
			case SDL_KEYDOWN:
				emu->out_debug(_T("SDL_KEYDOWN: scancode:%02x sym:%02x(%s) mod:%02x unicode:%04x")
					,e->key.keysym.scancode,e->key.keysym.sym,SDL_GetKeyName(e->key.keysym.sym),e->key.keysym.mod,e->key.keysym.unicode);

//				if (sym != SDLK_LALT && sym != SDLK_RALT && (mod & (KMOD_LALT | KMOD_RALT))) {
				if ((*key_mod) & KEY_MOD_ALT_KEY) {
					// translate sdl key to agar key
					// agar key is as same as sdl key.

					// execute for pressed global key
					if (gui->ExecuteGlobalKeys(sym, mod)) {
						// show menu bar
//						emu->out_debug("SDL_KEYDOWN: go to gui event loop.");
//						// mute sound
//						emu->mute_sound(true);
//						// stop key repeat
//						emu->key_lost_focus();
						//
						gui_process = true;
						gui->Show(gui_process);
					}
				} else {
					emu->key_down(sym, mod);
				}
				break;
			case SDL_KEYUP:
				emu->out_debug(_T("SDL_KEYUP: scancode:%02x sym:%02x(%s) mod:%02x unicode:%04x")
					,e->key.keysym.scancode,e->key.keysym.sym,SDL_GetKeyName(e->key.keysym.sym),e->key.keysym.mod,e->key.keysym.unicode);

				emu->key_up(sym, mod);
				break;
			}
		}
		break;
	case SDL_MOUSEMOTION:
		if (emu) {
			int x = ((SDL_MouseMotionEvent *)e)->x;
			int y = ((SDL_MouseMotionEvent *)e)->y;
			if (0 < y && y < 24 && 0 < x && x < SCREEN_WIDTH) {
				// show menu bar
//				emu->out_debug("SDL_MOUSEMOTION: y:%d -> Go AG ",((SDL_MouseMotionEvent *)e)->y);
//				// mute sound
//				emu->mute_sound(true);
//				// stop key repeat
//				emu->key_lost_focus();
				//
				gui_process = true;
				gui->Show(gui_process);
			}
		}
		break;
	case SDL_VIDEORESIZE:
		break;
	case SDL_QUIT:
		emu->out_debug(_T("SDL_QUIT"));
		working = false;
		break;
#ifdef USE_SOCKET
	case SDL_USEREVENT_SOCKET:
		if (emu) {
			int code = e->user.code;
			int no = (intptr_t)e->user.data1;
			switch(code) {
				case SDL_USEREVENT_SOCKET_CONNECTED:
					emu->socket_connected(no);
					break;
				case SDL_USEREVENT_SOCKET_DISCONNECTED:
					emu->socket_disconnected(no);
					break;
				case SDL_USEREVENT_SOCKET_WRITEABLE:
					emu->socket_writeable(no);
					emu->send_data(no);
					break;
				case SDL_USEREVENT_SOCKET_READABLE:
					emu->recv_data(no);
					emu->socket_readable(no);
					break;
				case SDL_USEREVENT_SOCKET_ACCEPT:
					emu->socket_accept(no);
					break;
			}
		}
		break;
#endif
	}
}


/*
 *	get and parse options
 */
static int get_options(int ac, char *av[])
{
	int opt = 0;

	memset(tape_file, '\0', _MAX_PATH);

	// set application path and name
	get_dir_and_basename(av[0], app_path, app_name);

#if !defined(_WIN32) || defined(__MINGW32__)
	while ((opt = getopt(ac, av, "ht:")) != -1) {
		switch (opt) {
		case 't':
			strncpy(tape_file, optarg, _MAX_PATH-1);
			break;
		case 'h':
			printf("Usage: %s [-h] [-t tape_image]\n",app_name);
			exit(0);
			break;
		default:
			break;
		}
	}
#endif
	return 0;
}

/*
 *	get dirname and basename
 */
void get_dir_and_basename(const char *path, char *dir, char *name)
{
	char *p = NULL;

	memset(dir, '\0', _MAX_PATH);
	if (name != NULL) memset(name, '\0', _MAX_PATH);
	strncpy(dir, path, _MAX_PATH-1);

#ifdef _WIN32
	p = strrchr(dir, '\\');
	if (p == NULL) {
#endif
		p = strrchr(dir, '/');
#ifdef _WIN32
	}
#endif
	if (p != NULL) {
		if (name != NULL) strncpy(name, p+1, _MAX_PATH-1);
		*(p+1) = '\0';
	} else {
		if (name != NULL) strncpy(name, dir, _MAX_PATH-1);
		*dir = '\0';
	}
	return;
}




