/** @file bml3mk5.cpp

	HITACHI BASIC MASTER LEVEL3 Mark5 Emulator 'EmuB-6892'
	Skelton for retropc emulator

	@author Sasaji
	@date   2011.06.08 -

	@brief [ virtual machine ]
*/

#include "bml3mk5.h"
#include "../../emu.h"
#include "../device.h"
#include "../event.h"

#include "../hd46505.h"
#include "../ym2203.h"
#include "../mc6809.h"

#include "board.h"
#include "pia.h"
#include "acia.h"
#include "display.h"
#include "keyboard.h"
#include "memory.h"
#include "sound.h"
#include "comm.h"
#include "cmt.h"
#include "printer.h"
#include "timer.h"
#include "kanji.h"
#include "psgc.h"
#include "via.h"
#include "psg9c.h"
#ifdef USE_FD1
#include "mc6843.h"
#include "mb8866.h"
#include "floppy.h"
#endif

// ----------------------------------------------------------------------------
// initialize
// ----------------------------------------------------------------------------

VM::VM(EMU* parent_emu) : emu(parent_emu)
{
	// create devices
	first_device = last_device = NULL;
	dummy = new DEVICE(this, emu, NULL);	// must be 1st device
	event = new EVENT(this, emu, NULL);	// must be 2nd device
	event->initialize();		// must be initialized first

	crtc = new HD46505(this, emu, NULL);
	psg[0] = new YM2203(this, emu, "P31");
	psg[1] = new YM2203(this, emu, "P32");
	psg9[0] = new YM2203(this, emu, "P91");
	psg9[1] = new YM2203(this, emu, "P92");
	psg9[2] = new YM2203(this, emu, "P93");
	cpu = new MC6809(this, emu, NULL);

	board = new BOARD(this, emu, NULL);
	pia = new PIA(this, emu, "MA");
	acia = new ACIA(this, emu, "MA");
	display = new DISPLAY(this, emu, NULL);
	key = new KEYBOARD(this, emu, NULL);
	memory = new MEMORY(this, emu, NULL);
	sound = new SOUND(this, emu, NULL);
	comm[0] = new COMM(this, emu, NULL, 0);
	comm[1] = new COMM(this, emu, NULL, 1);
	cmt = new CMT(this, emu, NULL);
	printer[0] = new PRINTER(this, emu, NULL, 0);
	printer[1] = new PRINTER(this, emu, NULL, 1);
	printer[2] = new PRINTER(this, emu, NULL, 2);
	timer = new TIMER(this, emu, NULL);
	kanji = new KANJI(this, emu, NULL);
	psg_pia = new PIA(this, emu, "P31");
	psgc = new PSGC(this, emu, "P31");
	psg9_via = new VIA(this, emu, "P91");
	psg9c = new PSG9C(this, emu, "P91");
#ifdef USE_FD1
	fdc3 = new MC6843(this, emu, NULL);
	fdc5 = new MB8866(this, emu, NULL);
	fdd = new FLOPPY(this, emu, NULL);
#endif
	pia_ex = new PIA(this, emu, "EX1");
	acia_ex = new ACIA(this, emu, "EX1");

	// set contexts
	event->set_context_cpu(cpu);
	event->set_context_sound(sound);
	event->set_context_sound(psg[0]);
	event->set_context_sound(psg[1]);
	event->set_context_sound(psg9[0]);
	event->set_context_sound(psg9[1]);
	event->set_context_sound(psg9[2]);
	event->set_context_sound(cmt);
	event->set_context_display(display);

	// crtc
	crtc->set_context_vsync(timer, SIG_TIMER_VSYNC, 1);
	crtc->set_context_vsync(display, SIG_DISPLAY_VSYNC, 1);
	crtc->set_context_hsync(key, SIG_KEYBOARD_HSYNC, 1);

	// display
	display->set_context_crtc(crtc);
	display->set_vram_ptr(memory->get_vram());
	display->set_color_ram_ptr(memory->get_color_ram());
	display->set_ig_ram_ptr(memory->get_ig_ram());
	display->set_regs_ptr(crtc->get_regs());
	display->set_vt_ptr(crtc->get_vt_total_ptr(), crtc->get_vt_count_ptr());

	// keyboard
	key->set_context_cpu(cpu);
	key->set_context_disp(display);
	key->set_context_board(board);

	// memory
	memory->set_context_cpu(cpu);
	memory->set_context_pia(pia);
	memory->set_context_acia(acia);
	memory->set_context_crtc(crtc);
	memory->set_context_display(display);
	memory->set_context_key(key);
	memory->set_context_sound(sound);
	memory->set_context_psg(psg_pia);
	memory->set_context_psg9(psg9_via);
	memory->set_context_comm(comm[0]);
	memory->set_context_cmt(cmt);
	memory->set_context_timer(timer);
	memory->set_context_kanji(kanji);
#ifdef USE_FD1
	memory->set_context_fdc(fdc3, fdc5);
	memory->set_context_fdd(fdd);
#endif
	memory->set_context_pia_ex(pia_ex);
	memory->set_context_acia_ex(acia_ex);
	memory->set_context_comm1(comm[1]);

	// cpu bus
	cpu->set_context_mem(memory);

	// acia
	acia->set_context_txdata(comm[0], SIG_ACIA_TXDATA, 0xff);
	acia->set_context_rts(comm[0], SIG_ACIA_RTS, 1);
	acia->set_context_dtr(comm[0], SIG_ACIA_DTR, 1);
	acia->set_context_irq(cpu, SIG_CPU_IRQ, 1);

	// acia_ex
	acia_ex->set_context_txdata(comm[1], SIG_ACIA_TXDATA, 0xff);
	acia_ex->set_context_rts(comm[1], SIG_ACIA_RTS, 1);
	acia_ex->set_context_dtr(comm[1], SIG_ACIA_DTR, 1);
	acia_ex->set_context_irq(cpu, SIG_CPU_IRQ, 1);

	// comm
	comm[0]->set_context_ctrl(acia);
	comm[0]->set_context_cmt(cmt);
	comm[1]->set_context_ctrl(acia_ex);

	// data recorder
	cmt->set_context_ctrl(comm[0]);

	// pia
	pia->set_context_pa(memory, SIG_MEMORY_PIA_PA, 0xff);
	pia->set_context_pb(printer[0], SIG_PRINTER_PIA_PB, 0xff);
	pia->set_context_cb2(printer[0], SIG_PRINTER_PIA_CB2, 1);

	// pia ex
	pia_ex->set_context_pa(printer[1], SIG_PRINTER_PIA_PB, 0xff);
	pia_ex->set_context_ca2(printer[1], SIG_PRINTER_PIA_CB2, 1);
	pia_ex->set_context_pb(printer[2], SIG_PRINTER_PIA_PB, 0xff);
	pia_ex->set_context_cb2(printer[2], SIG_PRINTER_PIA_CB2, 1);

	// printer
	printer[0]->set_context_ctrl(pia);
	printer[0]->set_context_cb1(SIG_PIA_CB1);
	printer[1]->set_context_ctrl(pia_ex);
	printer[1]->set_context_cb1(SIG_PIA_CA1);
	printer[2]->set_context_ctrl(pia_ex);
	printer[2]->set_context_cb1(SIG_PIA_CB1);

	// timer
	timer->set_context_cpu(cpu);

	// psg
	psg_pia->set_context_pa(psgc, SIG_PSG_PIA_PA, 0xff);
	psg_pia->set_context_pb(psgc, SIG_PSG_PIA_PB, 0xff);
	psgc->set_context_psg(psg[0], psg[1]);
	psgc->set_context_pia(psg_pia);

	// 9voice psg
	psg9_via->set_context_pa(psg9c, SIG_PSG9_VIA_PA, 0xff);
	psg9_via->set_context_pb(psg9c, SIG_PSG9_VIA_PB, 0xff);
	psg9_via->set_context_irq(cpu, SIG_CPU_IRQ, 1);
	psg9c->set_context_psg(psg9[0], psg9[1], psg9[2]);
	psg9c->set_context_via(psg9_via);

#ifdef USE_FD1
	// fdc for 3inch compact floppy
	fdc3->set_context_irq(cpu, SIG_CPU_NMI, 1);
	fdc3->set_context_fdd(fdd);

	// fdc for 5inch mini floppy
	fdc5->set_context_irq(fdd, SIG_FLOPPY_IRQ, 1);
	fdc5->set_context_drq(fdd, SIG_FLOPPY_DRQ, 1);
	fdc5->set_context_fdd(fdd);

	// fdd
	fdd->set_context_irq(cpu, SIG_CPU_NMI, 1);
//	fdd->set_context_drq(cpu, SIG_CPU_NMI, 1);
	fdd->set_context_fdc(fdc3, fdc5);
#endif

	// main board
	// reset signal
	board->set_context_reset(cpu, SIG_CPU_RESET, 1);
	board->set_context_reset(key, SIG_CPU_RESET, 1);
	board->set_context_reset(memory, SIG_CPU_RESET, 1);
	board->set_context_reset(cmt, SIG_CPU_RESET, 1);
	board->set_context_reset(pia, SIG_CPU_RESET, 1);
	board->set_context_reset(timer, SIG_CPU_RESET, 1);
	board->set_context_reset(psg[0], SIG_CPU_RESET, 1);
	board->set_context_reset(psg[1], SIG_CPU_RESET, 1);
	board->set_context_reset(psg_pia, SIG_CPU_RESET, 1);
	board->set_context_reset(psg9[0], SIG_CPU_RESET, 1);
	board->set_context_reset(psg9[1], SIG_CPU_RESET, 1);
	board->set_context_reset(psg9[2], SIG_CPU_RESET, 1);
	board->set_context_reset(psg9_via, SIG_CPU_RESET, 1);
#ifdef USE_FD1
	board->set_context_reset(fdd, SIG_CPU_RESET, 1);
#endif
	board->set_context_reset(pia_ex, SIG_CPU_RESET, 1);

	// initialize all devices
	for(DEVICE* device = first_device; device; device = device->next_device) {
		if(device->this_device_id != event->this_device_id) {
			device->initialize();
		}
	}
}

VM::~VM()
{
	// delete all devices
	for(DEVICE* device = first_device; device; device = device->next_device) {
		device->release();
	}
	for(DEVICE* device = first_device; device;) {
		DEVICE *next_device = device->next_device;
		delete device;
		device = next_device;
	}
}

DEVICE* VM::get_device(int id)
{
	for(DEVICE* device = first_device; device; device = device->next_device) {
		if(device->this_device_id == id) {
			return device;
		}
	}
	return NULL;
}

DEVICE* VM::get_device(char *name, char *identifier)
{
	for(DEVICE* device = first_device; device; device = device->next_device) {
		if(strncmp(name, device->this_class_name, 12) == 0
		&& strncmp(identifier, device->this_identifier, 4) == 0) {
			return device;
		}
	}
	return NULL;
}

// ----------------------------------------------------------------------------
// drive virtual machine
// ----------------------------------------------------------------------------

void VM::reset()
{
	// power on
	emu->out_info(_T("PowerOn"));

	// reset all devices
	for(DEVICE* device = first_device; device; device = device->next_device) {
		device->reset();
	}
	// set initial port status
	sound->write_signal(SIG_SOUND_ON, 0, 1);
}

void VM::special_reset()
{
}

void VM::warm_reset()
{
	// send reset signal
	board->write_signal(SIG_CPU_RESET, 1, 1);
	board->write_signal(SIG_CPU_RESET, 0, 1);
}

void VM::run(int split_num)
{
	event->drive(split_num);

}

double VM::frame_rate()
{
	return event->frame_rate();
}

void VM::change_dipswitch(int num)
{
	emu->out_infof(_T("Mode Switch %s"), (config.dipswitch & 4) ? "ON" : "OFF");
}

bool VM::now_skip()
{
	return false;
}

// ----------------------------------------------------------------------------
// event manager
// ----------------------------------------------------------------------------

void VM::register_event(DEVICE* dev, int event_id, int usec, bool loop, int* register_id)
{
	event->register_event(dev, event_id, usec, loop, register_id);
}

void VM::register_event_by_clock(DEVICE* dev, int event_id, int clock, bool loop, int* register_id)
{
	event->register_event_by_clock(dev, event_id, clock, loop, register_id);
}

void VM::cancel_event(int register_id)
{
	event->cancel_event(register_id);
}

void VM::register_frame_event(DEVICE* dev)
{
	event->register_frame_event(dev);
}

void VM::register_vline_event(DEVICE* dev)
{
	event->register_vline_event(dev);
}

uint64 VM::current_clock()
{
	return event->current_clock();
}

uint64 VM::passed_clock(uint64 prev)
{
	return event->passed_clock(prev);
}

//uint32 VM::get_pc()
//{
//	return cpu->get_pc();
//}

void VM::set_lines_per_frame(int lines) {
	event->set_lines_per_frame(lines);
}

// ----------------------------------------------------------------------------
// draw screen
// ----------------------------------------------------------------------------

void VM::draw_screen()
{
	display->set_vram_ptr(memory->get_vram());
	display->draw_screen();
}

// ----------------------------------------------------------------------------
int VM::update_led()
{
	int status = 0;

	status |= (key->get_kb_mode() & 0x07);
	status |= (cmt->get_cmt_mode() << 4);
#ifdef USE_FD1
	status |= (fdd->get_drive_select() << 8);
#endif
	return status;
}

// ----------------------------------------------------------------------------
// sound manager
// ----------------------------------------------------------------------------

void VM::initialize_sound(int rate, int samples)
{
	// init sound manager
	event->initialize_sound(rate, samples);

	// init sound gen
	sound->init(rate, 32768);
	psg[0]->init(rate, CPU_CLOCKS * 2, samples, 10, 19);
	psg[1]->init(rate, CPU_CLOCKS * 2, samples, 10, 19);
	psg9[0]->init(rate, CPU_CLOCKS * 4, samples, 10, 19);
	psg9[1]->init(rate, CPU_CLOCKS * 4, samples, 10, 19);
	psg9[2]->init(rate, CPU_CLOCKS * 4, samples, 10, 19);
	cmt->init(rate, 32767);

	//
	set_volume();
}

/// @attention called by another thread
#ifdef USE_AUDIO_U8
uint8* VM::create_sound(int* extra_frames, int samples)
#else
int16* VM::create_sound(int* extra_frames, int samples)
#endif
{
	return event->create_sound(extra_frames, samples);
}

void VM::set_volume()
{
	event->set_volume(config.mute ? 0 : config.volume);
	sound->set_volume(config.beep_mute ? 0 : config.beep_volume);
	psg[0]->set_volume(config.psg6_mute ? 0 : config.psg6_volume);
	psg[1]->set_volume(config.psg6_mute ? 0 : config.psg6_volume);
	psg9[0]->set_volume(config.psg9_mute ? 0 : config.psg9_volume);
	psg9[1]->set_volume(config.psg9_mute ? 0 : config.psg9_volume);
	psg9[2]->set_volume(config.psg9_mute ? 0 : config.psg9_volume);
	cmt->set_volume(config.relay_mute ? 0 : config.relay_volume, config.cmt_mute ? 0 : config.cmt_volume);
}

// ----------------------------------------------------------------------------
// notify key
// ----------------------------------------------------------------------------

void VM::key_down(int code)
{
}

void VM::key_up(int code)
{
}

// ----------------------------------------------------------------------------
// user interface
// ----------------------------------------------------------------------------

bool VM::play_datarec(_TCHAR* file_path)
{
	return cmt->play_datarec(file_path);
}

bool VM::rec_datarec(_TCHAR* file_path)
{
	return cmt->rec_datarec(file_path);
}

void VM::close_datarec()
{
	cmt->close_datarec();
}

void VM::rewind_datarec()
{
	cmt->rewind_datarec();
}

void VM::fast_forward_datarec()
{
	cmt->fast_forward_datarec();
}

void VM::realmode_datarec()
{
	cmt->realmode_datarec();
}

// ----------------------------------------------------------------------------

#ifdef USE_FD1
bool VM::open_disk(int drv, _TCHAR* file_path, int offset, uint32 flags)
{
	return fdd->open_disk(drv, file_path, offset, flags);
}

void VM::close_disk(int drv)
{
	fdd->close_disk(drv);
}

int VM::change_disk(int drv)
{
	switch(config.fdd_type) {
		case 1:
			return fdd->change_disk(drv);
			break;
	}
	return 0;
}

bool VM::disk_inserted(int drv)
{
	return fdd->disk_inserted(drv);
}

int VM::get_disk_side(int drv)
{
	return fdd->get_disk_side(drv);
}
#endif
// ----------------------------------------------------------------------------
void VM::update_config()
{
	for(DEVICE* device = first_device; device; device = device->next_device) {
		device->update_config();
	}
}

// ----------------------------------------------------------------------------
void VM::save_printer(int drv, _TCHAR* filename)
{
	printer[drv]->save_printer(filename);
}

void VM::clear_printer(int drv)
{
	printer[drv]->reset();
}

int VM::get_printer_buffer_size(int drv)
{
	return printer[drv]->get_buffer_size();
}

uint8* VM::get_printer_buffer(int drv)
{
	return printer[drv]->get_buffer();
}

void VM::enable_printer_direct(int drv)
{
	printer[drv]->set_direct_mode();
}

bool VM::print_printer(int drv)
{
	return printer[drv]->print_printer();
}

// ----------------------------------------------------------------------------
void VM::enable_comm_server(int drv)
{
	comm[drv]->enable_server();
}

void VM::enable_comm_connect(int drv)
{
	comm[drv]->enable_connect();
}

#if 0
void VM::recv_comm_data(uint8 *data, size_t size)
{
	comm->recv_data(data, size);
}
#endif

// ----------------------------------------------------------------------------
void VM::save_keybind()
{
	key->save_keybind();
}

// ----------------------------------------------------------------------------
void VM::save_state(_TCHAR* filename)
{
	FILEIO *fio = new FILEIO();
	vm_state_header_t vm_state_h;

	// header
	memset(&vm_state_h, 0, sizeof(vm_state_h));
	strncpy(vm_state_h.header, RESUME_FILE_HEADER, 16);
	vm_state_h.version = RESUME_FILE_VERSION;
	vm_state_h.revision = RESUME_FILE_REVISION;
	vm_state_h.param = 0;

	if(fio->Fopen(filename, FILEIO_WRITE_BINARY)) {
		// write header
		fio->Fwrite(&vm_state_h, sizeof(vm_state_h), 1);
		// write data
		for(DEVICE* device = first_device; device; device = device->next_device) {
			device->save_state(fio);
		}
		fio->Fclose();
	}

	delete fio;
}

void VM::load_state(_TCHAR* filename)
{
	FILEIO *fio = new FILEIO();
	vm_state_header_t vm_state_h;

	if(!fio->Fopen(filename, FILEIO_READ_BINARY)) {
		emu->out_log(LOG_ERROR, _T("Load State: Cannot open."));
		return;
	}
	// read header
	fio->Fread(&vm_state_h, sizeof(vm_state_h), 1);
	// check header
	if (strncmp(vm_state_h.header, RESUME_FILE_HEADER, 16) != 0) {
		emu->out_log(LOG_ERROR, _T("Load State: Invalid header."));
		fio->Fclose();
		return;
	}
	if (vm_state_h.version != RESUME_FILE_VERSION) {
		emu->out_log(LOG_ERROR, _T("Load State: Invalid version."));
		fio->Fclose();
		return;
	}
	// read data
	for(DEVICE* device = first_device; device; device = device->next_device) {
		device->load_state(fio);
	}
	fio->Fclose();

	delete fio;

}

// ----------------------------------------------------------------------------
// socket
// ----------------------------------------------------------------------------

void VM::network_connected(int ch)
{
	for(int i=0; i<MAX_PRINTER; i++) {
		printer[i]->network_connected(ch);
	}
	for(int i=0; i<MAX_COMM; i++) {
		comm[i]->network_connected(ch);
	}
}

void VM::network_disconnected(int ch)
{
	for(int i=0; i<MAX_PRINTER; i++) {
		printer[i]->network_disconnected(ch);
	}
	for(int i=0; i<MAX_COMM; i++) {
		comm[i]->network_disconnected(ch);
	}
}

void VM::network_writeable(int ch)
{
	for(int i=0; i<MAX_PRINTER; i++) {
		printer[i]->network_writeable(ch);
	}
	for(int i=0; i<MAX_COMM; i++) {
		comm[i]->network_writeable(ch);
	}
}

void VM::network_readable(int ch)
{
	for(int i=0; i<MAX_PRINTER; i++) {
		printer[i]->network_readable(ch);
	}
	for(int i=0; i<MAX_COMM; i++) {
		comm[i]->network_readable(ch);
	}
}

void VM::network_accepted(int ch, int new_ch)
{
	for(int i=0; i<MAX_COMM; i++) {
		comm[i]->network_accepted(ch, new_ch);
	}
}

uint8* VM::get_sendbuffer(int ch, int* size)
{
	uint8 *buf = NULL;
	for(int i=0; i<MAX_PRINTER; i++) {
		buf = printer[i]->get_sendbuffer(ch, size);
		if (buf != NULL) return buf;
	}
	for(int i=0; i<MAX_COMM; i++) {
		buf = comm[i]->get_sendbuffer(ch, size);
		if (buf != NULL) return buf;
	}
	return buf;
}

void VM::inc_sendbuffer_ptr(int ch, int size)
{
	for(int i=0; i<MAX_PRINTER; i++) {
		printer[i]->inc_sendbuffer_ptr(ch, size);
	}
	for(int i=0; i<MAX_COMM; i++) {
		comm[i]->inc_sendbuffer_ptr(ch, size);
	}
}

uint8* VM::get_recvbuffer0(int ch, int* size0, int* size1)
{
	uint8 *buf = NULL;
	for(int i=0; i<MAX_PRINTER; i++) {
		buf = printer[i]->get_recvbuffer0(ch, size0, size1);
		if (buf != NULL) return buf;
	}
	for(int i=0; i<MAX_COMM; i++) {
		buf = comm[i]->get_recvbuffer0(ch, size0, size1);
		if (buf != NULL) return buf;
	}
	return buf;
}

uint8* VM::get_recvbuffer1(int ch)
{
	uint8 *buf = NULL;
	for(int i=0; i<MAX_PRINTER; i++) {
		buf = printer[i]->get_recvbuffer1(ch);
		if (buf != NULL) return buf;
	}
	for(int i=0; i<MAX_COMM; i++) {
		buf = comm[i]->get_recvbuffer1(ch);
		if (buf != NULL) return buf;
	}
	return buf;
}

void VM::inc_recvbuffer_ptr(int ch, int size)
{
	for(int i=0; i<MAX_PRINTER; i++) {
		printer[i]->inc_recvbuffer_ptr(ch, size);
	}
	for(int i=0; i<MAX_COMM; i++) {
		comm[i]->inc_recvbuffer_ptr(ch, size);
	}
}
