/** @file display.cpp

	HITACHI BASIC MASTER LEVEL3 Mark5 Emulator 'EmuB-6892'
	Skelton for retropc emulator

	@author Sasaji
	@date   2011.06.08 -

	@brief [ display ]
*/

#include "display.h"
#include "../../fileio.h"
#include "../../config.h"

void DISPLAY::initialize()
{
	scanline = config.scan_line;
	afterimage = config.afterimage;
	keepimage = config.keepimage;

	memset(font, 0, sizeof(font));

	// load rom images
	_TCHAR *app_path, *rom_path[2], file_path[_MAX_PATH];
	bool rom_loaded = false;

	rom_path[0] = config.rom_path;
	rom_path[1] = emu->application_path();

	FILEIO* fio = new FILEIO();
	for(int i=0; i<2; i++) {
		app_path = rom_path[i];

		if (!rom_loaded) {
			_stprintf(file_path, _T("%sFONT.ROM"), app_path);
			if(fio->Fopen(file_path, FILEIO_READ_BINARY)) {
				fio->Fread(font, sizeof(font), 1);
				fio->Fclose();
				rom_loaded = true;
				emu->out_log(LOG_INFO, _T("FONT.ROM was loaded."));
			}
		}
	}
	delete fio;

	if (!rom_loaded) {
		emu->out_log(LOG_WARN, _T("FONT.ROM couldn't be loaded."));
	}

	emu->get_rgbformat(&Rmask, &Gmask, &Bmask, &Rshift, &Gshift, &Bshift);

	// create pc palette
	for(int i = 0; i < 8; i++) {
//		palette_pc[i] = RGB_COLOR((i & 2) ? 255 : 0, (i & 4) ? 255 : 0, (i & 1) ? 255 : 0);
		palette_pc[i] = emu->map_rgbcolor((i & 2) ? 255 : 0, (i & 4) ? 255 : 0, (i & 1) ? 255 : 0);
		emu->out_debug(_T("palette_pc[%d]:%x"),i,palette_pc[i]);
	}
	for(int y=0; y<SCREEN_HEIGHT; y++) {
		for(int x=0; x<SCREEN_WIDTH; x++) {
			scrntop[y][x] = 0;
		}
	}

	mode_sel = 0;
	ig_enable = 0;
	ig_mask = true;

	start_addr = 0x4000;
	cursor_addr = 0;
	cblink = 0;
	disp_cursor = false;
	for(int i=0; i<32; i++) {
		cur_left[i] = 0;
		cur_right[i] = 0;
	}
	cur_top = 0;
	cur_bottom = 0;
	interace = 0;
	videomode = 0;
	disptmg_skew = 0;
	curdisp_skew = 0;

	bg_color = 0;
	hireso = 0;
	width = 1;
	width_sel = 0;

	top_raster = 0;
	bottom_raster = 0;
	vsync_top_raster = 0;
	vsync_bottom_raster = 0;
	raster_per_char = 16;
	now_raster = 0;
	lines_per_2frames = (LINES_PER_FRAME << 1);

	// synchronizable vertical range
	v_total_min[0] = (LINES_PER_FRAME - 24);
	v_total_max[0] = (LINES_PER_FRAME + 24);
	v_total_min[1] = (LINES_PER_FRAME - 24) / 2;
	v_total_max[1] = (LINES_PER_FRAME + 24) / 2;
	v_total_min[2] = (LINES_PER_FRAME - 24) * 2;
	v_total_max[2] = (LINES_PER_FRAME + 24) * 2;

	for(int i=0; i<VRAM_BUF_HEIGHT; i++) {
		left_col[i] = 0;
		right_col[i] = 0;
		hsync_left_col[i] = 0;
		hsync_right_col[i] = 0;
	}
	left_diff = 0;
	// synchronizable horizontal range
	h_total_min[0] = CHARS_PER_LINE - 3;	// w.40
	h_total_max[0] = CHARS_PER_LINE + 3;
	h_total_min[1] = CHARS_PER_LINE * 2 - 4;	// w.80
	h_total_max[1] = CHARS_PER_LINE * 2 + 4;

	sw = 0;

	Rloss = (64 << Rshift);
	Gloss = (32 << Gshift);
	Bloss = (96 << Bshift);

	for(int i=0; i<8; i++) {
		ppd[i]=(mode_sel & 0x80) ? 1 : 2;	// pixel per 1dot
		dof[i]=(0x80 >> i);
	}
	// pixel per 1char
	ppch = (mode_sel & 0x80) ? 8 : 16;

	//
	memset(vram_buf, 0, sizeof(vram_buf));
	memset(attr_buf, 0, sizeof(attr_buf));
	buf_col = 0;
	buf_row = 0;
	buf_addr = ADDR_VRAM_START;
	buf_ras = 0;
	buf_page = 0;

	cur_buf_col = 0;
	cur_buf_row = 0;

	skip_frame_count = 0;
	skip_frame = emu->get_skip_frame();

	// register event
	register_frame_event(this);
}

void DISPLAY::reset()
{
	ig_mask = true;
}

void DISPLAY::update_config()
{
	scanline = config.scan_line;
	afterimage = config.afterimage;
	keepimage = config.keepimage;
	if (keepimage > 0) {
		for(int y=0; y<SCREEN_HEIGHT; y++) {
			for(int x=0; x<SCREEN_WIDTH; x++) {
				scrntop[y][x] = 0;
			}
		}
	}
}

void DISPLAY::write_io8(uint32 addr, uint32 data)
{
	switch(addr & 0xffff) {
	case 0xffc6:
		// crtc addr register
		crtc_addr = (data & 0xff);
		break;
	case 0xffc7:
		// crtc data register
		switch(crtc_addr) {
			case 1:
				width = regs[1] ? regs[1] : 1;					// width
				break;
			case 8:
				interace = (regs[8] & 1) ? 1 : 0;				// 0:noninterace 1:interace
				videomode = ((regs[8] & 3) == 3) ? 1 : 0;		// 1:videomode(interace only)

				disptmg_skew = ((regs[8] & 0x30) >> 4);			// disptmg skew
				curdisp_skew = ((regs[8] & 0xc0) >> 6);			// curdisp skew
//				break;	// through
			case 9:
				// raster of 1char
				raster_per_char = (regs[9] & 0x1f) + videomode + 1;
				raster_per_char = raster_per_char * (2 - videomode);
				break;

		}
		break;
	case 0xffd0:
		// mode sel
		mode_sel = (data & 0xff);

		bg_color = mode_sel & 0x07;				// bg color
		hireso = (mode_sel & 0x40) ? 0 : 1;		// 0:normal 1:hireso
		width_sel = (mode_sel & 0x80) ? 1 : 0;	// 0:w.40  1:w.80

		for(int i=0; i<8; i++) {
			ppd[i]=width_sel ? 1 : 2;	// pixel per 1dot
			dof[i]=(0x80 >> i);
		}
		// pixel per 1char
		ppch = (mode_sel & 0x80) ? 8 : 16;

		break;
	case 0xffd6:
		// interace sel
		interace_sel = (data & 0x08) ? 1 : 0;	// 0:noninterace char  1:interace char
		break;
	case 0xffe9:
		// ig mode reg
		ig_enable = (data & 0x01);
		ig_mask = false;
		break;
	}
}

void DISPLAY::write_signal(int id, uint32 data, uint32 mask)
{
	switch(id) {
		case SIG_DISPLAY_LPSTB:
			{
			int width = regs[1];
			int dotx = (data >> 16) & 0xffff;
			int doty = (data & 0xffff);
			int x = 0;
			int y = 0;
			int offset = (width_sel ? 5 : 4);
			uint32 newaddr;

			if (dotx < left_col[doty >> 1]) { dotx = right_col[doty >> 1] - 1; doty-=16; }
			if (dotx >= right_col[doty >> 1]) { dotx = left_col[doty >> 1]; doty+=16; }
			if (doty < top_raster) doty = top_raster;
			if (doty >= bottom_raster) doty = bottom_raster - 1;

			x = (dotx - left_col[doty >> 1]) >> (width_sel ? 3 : 4);
			y = (doty - top_raster) >> 4;
			newaddr = ((regs[12] << 8)+regs[13]) + y * width + x + offset;
			regs[16] = (newaddr >> 8) & 0x3f;
			regs[17] = newaddr & 0xff;

//			emu->out_debug("lp dx:%03d dy:%03d x:%02d y:%02d addr:%02x%02x"
//				,dotx,doty,x,y,regs[16],regs[17]);
			}
			break;

		case SIG_DISPLAY_VSYNC:
			if ((data & mask) == 0) {
				// synchronize display vertical raster
				for(int i=0; i<3; i++) {
					if ((v_total_min[i] <= now_raster && now_raster <= v_total_max[i])
					|| (v_total_min[i] <= *vt_total && *vt_total <= v_total_max[i])) {
						now_raster = 0;
						break;
					}
				}
				// cursor blink
				cblink = (cblink + 1) & 0x1f;

//				emu->out_debug("vsync now:%d vt_total:%d vt_count:%d"
//				,now_raster,*vt_total,*vt_count);
			}
			break;
	}
}

void DISPLAY::update_cursor()
{
	int bp = regs[10] & 0x60;						// cursor blink speed

	uint8 cur_st_l;
	uint8 cur_ed_l;

	if(cur_buf_col < 0 || cur_buf_row < 0 || curdisp_skew == 3) {
		disp_cursor = false;
	} else {
		disp_cursor = (bp == 0 || (bp == 0x40 && (cblink & 8)) || (bp == 0x60 && (cblink & 0x10)));
	}

	if (disp_cursor) {
		if (videomode == 1) {
			cur_st_l = (regs[10] & 0x0f);
			cur_ed_l = (regs[11] & 0x0f);
		} else {
			cur_st_l = ((regs[10] & 0x0f) << 1);
			cur_ed_l = ((regs[11] & 0x0f) << 1) + 1;
		}

		// cursor draw area
		cur_top = top_raster + cur_buf_row * 2;
		cur_bottom = cur_top + cur_ed_l;
		cur_top += cur_st_l;

		for(int i=0; i < raster_per_char; i++) {
			cur_left[i] = left_col[cur_buf_row] + (cur_buf_col + curdisp_skew + config.curdisp_skew - 2 - disptmg_skew + config.disptmg_skew) * ppch;
			cur_right[i] = cur_left[i] + ppch;
		}
	}
	cur_buf_col = -1;
	cur_buf_row = -1;
}

void DISPLAY::draw_screen()
{
	if (*skip_frame) {
		skip_frame_count++;
	} else {
		skip_frame_count = 0;
	}
	if ((afterimage == 0 && skip_frame_count >= 1)
	||  (afterimage > 0 && skip_frame_count >= 2)) {
		// clear cursor
		disp_cursor = false;
		cur_buf_col = -1;
		cur_buf_row = -1;

		return;
	}

	sw = 1 - sw;

	// update cursor
	update_cursor();

	// render screen
	draw_screen_sub();
}

uint8 DISPLAY::get_font_data(uint8 chr_code, uint8 y)
{
	uint16 code = 0;
	uint8  offset = 0;
	if (interace_sel == 1) {
		// interace mode char
		if (videomode == 0) {
			y = y >> 1;	//	double size
		}
		if (chr_code < 0x80) {
			code = chr_code << 4;	// 16byte per 1char
			offset = (y & 0x0e);	// 0 0 2 2 4 4 6 6 8 8 a a c c e e
		} else {
			code = chr_code << 4;	// 16byte per 1char
			offset = (y & 0x0f);	// 0 1 2 3 4 5 6 7 8 9 a b c d e f
		}
	} else {
		// non interace mode char
		if (chr_code < 0x80) {
			// ascii code
			code = chr_code << 4;	// 16byte per 1char
		} else {
			code = ((chr_code - 0x80) << 4) + 1;
		}
		if (videomode == 0) {
			offset = (y & 0x0e);		// 0 0 2 2 4 4 6 6 8 8 a a c c e e
		} else {
			offset = ((y << 1) & 0x0e);	// 0 2 4 6 8 a c e 0 2 4 6 8 a c e
		}
	}
	return font[code + offset];
}

#define GET_FONT_DATA(chr_code, y, new_code) { \
	uint16 new_chr_code = 0; \
	uint8  offset = 0; \
	int yy = y; \
	if (interace_sel == 1) { \
		if (videomode == 0) { \
			yy = yy >> 1; \
		} \
		if (chr_code < 0x80) { \
			new_chr_code = chr_code << 4; \
			offset = (yy & 0x0e); \
		} else { \
			new_chr_code = chr_code << 4; \
			offset = (yy & 0x0f); \
		} \
	} else { \
		if (chr_code < 0x80) { \
			new_chr_code = chr_code << 4; \
		} else { \
			new_chr_code = ((chr_code - 0x80) << 4) + 1; \
		} \
		if (videomode == 0) { \
			offset = (yy & 0x0e); \
		} else { \
			offset = ((yy << 1) & 0x0e); \
		} \
	} \
	new_code = font[new_chr_code + offset]; \
}

#define COLOR_REVERSE(fg,bg) { \
	{ \
		uint8 c; \
		c = (bg); \
		(bg) = (fg); \
		(fg) = c; \
	} \
}

void DISPLAY::draw_screen_sub()
{
	uint8 p;
	uint8 code;
	uint8 c_f;
	uint8 c_b;
	uint8 ck_f;
	uint8 ck_b;
	uint8 cram = 0;
	uint8 i_r,i_g,i_b;
	int bt;
	int bi;
	int dt, dtm;
	uint8 fg;
	int x,y;
	int l = top_raster <= bottom_raster ? 0 : raster_per_char - ((bottom_raster - 1) % raster_per_char) - 1;
	bool disp_cursor_line = false;
	bool disp_cursor_char = false;
	bool ki;
	bool skip_line;
	int buf_x = 0;
	int buf_y = - top_raster + (VRAM_BUF_HEIGHT << 1);
	int buf_y_h;

	int scrn_offset = emu->screen_buffer_offset();
	scrntype *scrn = emu->screen_buffer(0);
	scrntype dot = 0;
	uint32 r,g,b;

	for(y = 0; y < SCREEN_HEIGHT; ) {
		buf_y = buf_y % (VRAM_BUF_HEIGHT << 1);
		buf_y_h = buf_y >> 1;

		skip_line = scanline && ((interace == 0 && (y & 1) != 0) || (interace == 1 && sw != (y & 1)));

		// 
		if (afterimage == 0) {
			for(x = 0; x < SCREEN_WIDTH; x++) {
				scrnline1[x] = 0;
				scrnline2[x] = 0xffffffff;
			}
		} else if (afterimage == 1) {
			for(x = 0; x < SCREEN_WIDTH; x++) {
				dot =scrn[x];
				r = (dot & Rmask);
				g = (dot & Gmask);
				b = (dot & Bmask);

				if (r > Rloss) r = r - Rloss;
				else r = 0;
				if (g > Gloss) g = g - Gloss;
				else g = 0;
				if (b > Bloss) b = b - Bloss;
				else b = 0;

				dot = r | g | b;

				scrnline1[x] = dot;
				scrnline2[x] = 0xffffffff;
			}
		} else if (afterimage == 2) {
			for(x = 0; x < SCREEN_WIDTH; x++) {
				// 0xff or 0xdf -> 0xe0 -> 0x00
//				dot = scrn[x] & 0x10101010;
//				dot = (dot << 3) | (dot << 2) | (dot << 1);
//				scrnline2[x] = 0xdfdfdfdf;
				// 0xff or 0xbf -> 0xc0 -> 0x00
				dot = scrn[x] & 0x20202020;
				dot = (dot << 2) | (dot << 1);
				scrnline2[x] = 0xbfbfbfbf;

				scrnline1[x] = dot;
			}
		}

		if ((disptmg_skew == 3)
		||  (top_raster <  bottom_raster && (y < top_raster || bottom_raster <= y))
		||  (top_raster >= bottom_raster && (bottom_raster < y && y <= top_raster))) {
			// out of disptmg area
			if ((skip_line)
			||  (vsync_top_raster <= y && y < vsync_bottom_raster)) {
				for(x = 0; x < SCREEN_WIDTH; x++) {
					scrn[x] = (palette_pc[0] & scrnline2[x]) | scrnline1[x];
				}
			} else {
				for(x = 0; x < SCREEN_WIDTH; x++) {
					if (hsync_left_col[buf_y_h] <= x && x < hsync_right_col[buf_y_h]) {
						scrn[x] = (palette_pc[0] & scrnline2[x]) | scrnline1[x];
					} else {
						scrn[x] = (palette_pc[bg_color] & scrnline2[x]) | scrnline1[x];
					}
				}
			}
			y++;
			buf_y++;
			scrn += scrn_offset;
			continue;
		}

		// disptmg area
		if (!skip_line) {
			buf_x = left_col[buf_y_h] < 0 ? - left_col[buf_y_h] / ppch : 0 ;
			disp_cursor_line = (disp_cursor && cur_top <= y && y <= cur_bottom);
			for(x = 0; x < SCREEN_WIDTH; ) {
				disp_cursor_char = (disp_cursor_line && cur_left[l] <= x && x < cur_right[l]);

				// out of disptmg area
				if (x < left_col[buf_y_h] || right_col[buf_y_h] <= x) {
					if (hsync_left_col[buf_y_h] <= x && x < hsync_right_col[buf_y_h]) {
						c_f = 0;
						c_b = 0;
					} else {
						c_f = cram & 0x07;
						c_b = bg_color;
					}
					if (disp_cursor_char) COLOR_REVERSE(c_f, c_b);
					scrn[x] = (palette_pc[c_b] & scrnline2[x]) | scrnline1[x];
					x++;
					continue;
				}

				code = vram_buf[buf_y_h][buf_x];
				cram = attr_buf[buf_y_h][buf_x];
				ki = cram & 0x80 ? true : false;
				c_f = cram & 0x07;
				c_b = bg_color;
				ck_f = c_f;
				ck_b = 0;
				// color reverse
				if ((cram & 0x28) == 0x08) {
					COLOR_REVERSE(c_f, c_b);
					COLOR_REVERSE(ck_f, ck_b);
				}

				if (cram & 0x20) {
					// ig
					if (!ig_enable) {
						// disp ig pattern
						int ll = l >> ((interace_sel) + (1 - videomode));
						i_b = ig_ram[code*8+ll];
						i_r = ig_ram[code*8+ll+(ADDR_IGRAM_END-ADDR_IGRAM_START)];
						i_g = ig_ram[code*8+ll+(ADDR_IGRAM_END-ADDR_IGRAM_START)*2];

						bi = 0;
						for(bt = 0; bt < 8; bt++) {
							p = ((i_g & dof[bt]) ? 4 : 0) | ((i_r & dof[bt]) ? 2 : 0) | ((i_b & dof[bt]) ? 1 : 0);
							if (p == 0) p = c_b;
							dtm = x+bi+ppd[bt];
							if (dtm > SCREEN_WIDTH) dtm = SCREEN_WIDTH;
							for(dt = x+bi; dt < dtm; dt++) {
								scrn[dt] = (palette_pc[(p > 0) ? p : c_b] & scrnline2[dt]) | scrnline1[dt];
							}
							bi += ppd[bt];
						}
						if (ki) {
							bi = 0;
							for(bt = 0; bt < 8; bt++) {
								p = ((i_g & dof[bt]) ? 4 : 0) | ((i_r & dof[bt]) ? 2 : 0) | ((i_b & dof[bt]) ? 1 : 0);
								dtm = x+bi+ppd[bt];
								if (dtm > SCREEN_WIDTH) dtm = SCREEN_WIDTH;
								for(dt = x+bi; dt < dtm; dt++) {
									scrntop[y][dt] = palette_pc[p];
								}
								bi += ppd[bt];
							}
						}
					} else {
						// disp white square
						p = 0xff;
						c_f = 7;
						c_b = bg_color;
						ck_f = 7;
						ck_b = 0;

						if (disp_cursor_char) COLOR_REVERSE(c_f, c_b);

						dtm = x+ppch;
						if (dtm > SCREEN_WIDTH) dtm = SCREEN_WIDTH;
						for(dt = x; dt < dtm; dt++) {
							scrn[dt] = (palette_pc[c_f] & scrnline2[dt]) | scrnline1[dt];
						}
						if (ki) {
							for(dt = x; dt < dtm; dt++) {
								scrntop[y][dt] = palette_pc[ck_f];
							}
						}
					}
				} else if (cram & 0x10) {
					// graphic 
					p = code;

					if (disp_cursor_char) COLOR_REVERSE(c_f, c_b);
					if (cram & 0x40) {
						// hireso graphic
						bi = 0;
						for(bt = 0; bt < 8; bt++) {
							fg = (p & dof[bt]) ? c_f : c_b;
							dtm = x+bi+ppd[bt];
							if (dtm > SCREEN_WIDTH) dtm = SCREEN_WIDTH;
							for(dt = x+bi; dt < dtm; dt++) {
								scrn[dt] = (palette_pc[fg] & scrnline2[dt]) | scrnline1[dt];
							}
							bi += ppd[bt];
						}
						if (ki) {
							bi = 0;
							for(bt = 0; bt < 8; bt++) {
								fg = (p & dof[bt]) ? ck_f : ck_b;
								dtm = x+bi+ppd[bt];
								if (dtm > SCREEN_WIDTH) dtm = SCREEN_WIDTH;
								for(dt = x+bi; dt < dtm; dt++) {
									scrntop[y][dt] = palette_pc[fg];
								}
								bi += ppd[bt];
							}
						}
					} else {
						// normal graphic
						bi = 0;
						for (bt = 0; bt < 8; bt += 4) {
							fg = (p & ((0x10 >> bt) << (l >> 2))) ? c_f : c_b;
							dtm = x+bi+ppd[bt]+ppd[bt+1]+ppd[bt+2]+ppd[bt+3];
							if (dtm > SCREEN_WIDTH) dtm = SCREEN_WIDTH;
							for(dt = x+bi; dt < dtm; dt++) {
								scrn[dt] = (palette_pc[fg] & scrnline2[dt]) | scrnline1[dt];
							}
							bi += ppd[bt] + ppd[bt + 1] + ppd[bt + 2] + ppd[bt + 3];
						}
						if (ki) {
							bi = 0;
							for (bt = 0; bt < 8; bt += 4) {
								fg = (p & ((0x10 >> bt) << (l >> 2))) ? ck_f : ck_b;
								dtm = x+bi+ppd[bt]+ppd[bt+1]+ppd[bt+2]+ppd[bt+3];
								if (dtm > SCREEN_WIDTH) dtm = SCREEN_WIDTH;
								for(dt = x+bi; dt < dtm; dt++) {
									scrntop[y][dt] = palette_pc[fg];
								}
								bi += ppd[bt] + ppd[bt + 1] + ppd[bt + 2] + ppd[bt + 3];
							}
						}
					}
				} else {
					// character
//					p = get_font_data(code, l);
					GET_FONT_DATA(code, l, p);

					if (disp_cursor_char) COLOR_REVERSE(c_f, c_b);
					bi = 0;
					for(bt = 0; bt < 8; bt++) {
						fg = (p & dof[bt]) ? c_f : c_b;
						dtm = x+bi+ppd[bt];
						if (dtm > SCREEN_WIDTH) dtm = SCREEN_WIDTH;
						for(dt = x+bi; dt < dtm; dt++) {
							scrn[dt] = (palette_pc[fg] & scrnline2[dt]) | scrnline1[dt];
						}
						bi += ppd[bt];
					}
					if (ki) {
						bi = 0;
						for(bt = 0; bt < 8; bt++) {
							fg = (p & dof[bt]) ? ck_f : ck_b;
							dtm = x+bi+ppd[bt];
							if (dtm > SCREEN_WIDTH) dtm = SCREEN_WIDTH;
							for(dt = x+bi; dt < dtm; dt++) {
								scrntop[y][dt] = palette_pc[fg];
							}
							bi += ppd[bt];
						}
					}
				}
				x += ppch;
				buf_x++;
			}
		} else {
			// skip line
			for(x = 0; x < SCREEN_WIDTH; x++) {
				scrn[x] = (palette_pc[0] & scrnline2[x]) | scrnline1[x];
			}
		}

		y++;
		buf_y++;
		scrn += scrn_offset;
		l = ((l + 1) % raster_per_char);
	}

	// overlap the top most screen if need
	if (keepimage > 0) {
		scrn = emu->screen_buffer(0);
		for(y = 0; y < SCREEN_HEIGHT; y++) {
			for(x = 0; x < SCREEN_WIDTH; x++) {
				if (scrntop[y][x] != palette_pc[0]) scrn[x] = scrntop[y][x];
			}
			scrn += scrn_offset;
		}
	}
}

void DISPLAY::update_vram(int v, int clock)
{
	if (*vt_count != buf_row) {
		if (*vt_count == 0) {
			// disptmg on

			// start addr
			start_addr = (((regs[12] << 8) | regs[13]) + disptmg_skew - config.disptmg_skew) & 0x3fff;
			if (start_addr < ADDR_VRAM_START) {
				start_addr += (ADDR_VRAM_END - ADDR_VRAM_START);
			}

			// pixel of height
			top_raster = now_raster * 2;
			top_raster += (SCREEN_HEIGHT - 488) / 2;
			bottom_raster = top_raster + (regs[6] * raster_per_char);

			vsync_top_raster = top_raster + (regs[7] * raster_per_char);
			vsync_bottom_raster = vsync_top_raster + (((regs[3] & 0xf0) ? (regs[3] >> 4) : 16) << 1);

//			top_raster = (top_raster + lines_per_2frames) % lines_per_2frames;
//			bottom_raster = (bottom_raster + lines_per_2frames) % lines_per_2frames;

//			vsync_top_raster = (vsync_top_raster + lines_per_2frames) % lines_per_2frames;
//			vsync_bottom_raster = (vsync_bottom_raster + lines_per_2frames) % lines_per_2frames;

			// synchronize horizontal total
			if (h_total_min[width_sel] <= regs[0] && regs[0] <= h_total_max[width_sel]) {
				left_diff = 0;
			}
			// top most page ?
			buf_page = ((keepimage == 1 && start_addr < 0x0800) || (keepimage == 2 && start_addr >= 0x0800)) ? 0x80 : 0;
		}

		buf_row = *vt_count;
		buf_col = 0;
		buf_ras = (buf_row % (raster_per_char >> 1));
		buf_addr = start_addr + (buf_row / (raster_per_char >> 1)) * width;
		buf_addr += hireso * (buf_ras % 8) * (0x400 << width_sel);

		// cursor
		cursor_addr = (((regs[14] << 8) | regs[15]) & 0x3fff) + disptmg_skew - config.disptmg_skew;
		if (cursor_addr < 0x400) cursor_addr += 0x4000;

		// pixel of width
		if (buf_row < VRAM_BUF_HEIGHT) {
			left_col[buf_row] = regs[0]-regs[2]-(regs[3] & 0x0f)-(width_sel ? 14 : 5) + disptmg_skew - config.disptmg_skew;
			left_col[buf_row] *= ppch;
			left_col[buf_row] += (SCREEN_WIDTH - 768) / 2;
			left_col[buf_row] += left_diff;

			right_col[buf_row] = left_col[buf_row] + ppch * width;

			hsync_left_col[buf_row] = left_col[buf_row] + regs[2] * ppch;
			hsync_right_col[buf_row] = hsync_left_col[buf_row] + (regs[3] & 0x0f) * ppch;
		}
		// diff horizontal total display and crtc
		if (regs[0] < h_total_min[width_sel] || h_total_max[width_sel] < regs[0]) {
			left_diff += (regs[0] + 1 - (1 << (width_sel + 6))) * ppch;

			if (left_diff >= SCREEN_WIDTH) left_diff = -SCREEN_WIDTH + 1;
			else if (left_diff <= -SCREEN_WIDTH) left_diff = SCREEN_WIDTH - 1;
		}

		now_raster++;
		if (now_raster >= lines_per_2frames) now_raster = 0;
	}

	// copy vram data to buffer
	for(int i=0; i < (clock << width_sel); i++) {
		if (buf_addr >= ADDR_VRAM_END) {
			buf_addr -= (ADDR_VRAM_END - ADDR_VRAM_START);
		}
		if (buf_row < VRAM_BUF_HEIGHT) {
			if (buf_col < width) {
				vram_buf[buf_row][buf_col] = vram[buf_addr];
				// bit 0-5: color ram   bit 6: hireso  bit 7:top_most
				attr_buf[buf_row][buf_col] = (color_ram[buf_addr - ADDR_VRAM_START] | (hireso << 6) | buf_page) & (ig_mask ? 0xdf : 0xff);
				// cursor
				if (buf_addr == cursor_addr && buf_ras == 0) {
					cur_buf_col = buf_col;
					cur_buf_row = buf_row;
				}
			} else if (buf_col < VRAM_BUF_WIDTH) {
				vram_buf[buf_row][buf_col] = 0;
				attr_buf[buf_row][buf_col] = 0;
			}
		}
		buf_col++;
		buf_addr++;
	}
}

// ----------------------------------------------------------------------------
// event handler
// ----------------------------------------------------------------------------

void DISPLAY::event_frame()
{
//	cblink = (cblink + (1 << config.cpu_power)) & 0x1f;
//	sw = 1 - sw;
}

// ----------------------------------------------------------------------------

void DISPLAY::save_state(FILEIO *fio)
{
	struct vm_state_st vm_state;

	//
	vm_state_ident.version = 1;
	vm_state_ident.size = sizeof(vm_state_ident) + sizeof(vm_state);

	// copy values
	memset(&vm_state, 0, sizeof(vm_state));
	vm_state.crtc_addr = crtc_addr;
	vm_state.ig_enable = ig_enable;
	vm_state.ig_mask = ig_mask ? 1 : 0;

	fio->Fwrite(&vm_state_ident, sizeof(vm_state_ident), 1);
	fio->Fwrite(&vm_state, sizeof(vm_state), 1);
}

void DISPLAY::load_state(FILEIO *fio)
{
	vm_state_ident_t vm_state_i;
	struct vm_state_st vm_state;

	READ_STATE_CHUNK(fio, vm_state_i, vm_state);

	// require CRTC register
	for(crtc_addr = 0; crtc_addr < 16; crtc_addr++) {
		write_io8(0xffc7, regs[crtc_addr]);
	}

	crtc_addr = vm_state.crtc_addr;
	ig_enable = vm_state.ig_enable;
	ig_mask = vm_state.ig_mask ? true : false;

	// clear
	buf_row = 9999;
	buf_col = 0;

}

