/** @file memory.h

	HITACHI BASIC MASTER LEVEL3 Mark5 Emulator 'EmuB-6892'
	Skelton for retropc emulator

	@author Sasaji
	@date   2011.06.08 -

	@brief [ memory ]
*/

#ifndef _MEMORY_H_
#define _MEMORY_H_

#include "../vm.h"
#include "../../emu.h"
#include "../device.h"
#include "../../config.h"
#include "../hd46505.h"
#include "sound.h"
#include "display.h"
#include "comm.h"
#include "cmt.h"
#include "pia.h"
#include "via.h"

#define SIG_MEMORY_DISP		0
#define SIG_MEMORY_VSYNC	1
#define SIG_MEMORY_PIA_PA	2

#define BANK_SIZE	4

/**
	@brief Memory access
*/
class MEMORY : public DEVICE
{
private:
	DEVICE *d_cpu, *d_pia, *d_crtc, *d_disp, *d_acia;
	DEVICE *d_key, *d_sound, *d_comm, *d_cmt, *d_timer;
	DEVICE *d_kanji, *d_fdc3, *d_fdc5, *d_fdd, *d_psg, *d_psg9;
	DEVICE *d_pia_ex, *d_acia_ex, *d_comm1;

	uint8 rom[0x6000];
	uint8 rom1802[0x800];
	uint8 rom1805[0x800];
	uint8 ram[0x10000];
	uint8 color_ram[0x4000];	// 0x0400 - 0x43ff
	uint8 ig_ram[0x0800 * 3];	// 0xa000 - 0xa7ff * 3 (b,r,g)
	uint8 ram2[0x10000];		// extended ram

	uint8 wdmy[1 << BANK_SIZE];
	uint8 rdmy[1 << BANK_SIZE];
	uint8* wbank[0x10000 >> BANK_SIZE];
	uint8* rbank[0x10000 >> BANK_SIZE];

	uint8 color_reg;

	uint8 ig_enable;
	uint8 ig_enreg;

	uint8 mem_bank_reg1;	// from PIA port A
	uint8 mem_bank_reg2;	// from 0xffe8

	uint8 vram_sel;

	int   trace_counter_id;
	uint32 pc, pc_prev;

	bool  rom_loaded;
	bool  rom1802_loaded;
	bool  rom1805_loaded;

	//for resume
#pragma pack(1)
	struct vm_state_st {
		uint8 ram[0x10000];
		uint8 ram2[0x10000];		// extended ram
		uint8 color_ram[0x4000];	// 0x0400 - 0x43ff
		uint8 ig_ram[0x0800 * 3];	// 0xa000 - 0xa7ff * 3 (b,r,g)

		uint8 mem_bank_reg1;	// from PIA port A
		uint8 mem_bank_reg2;	// from 0xffe8

		int   trace_counter_id;
		uint32 pc;
		uint32 pc_prev;

		// from config
		uint8 dipswitch;
		int fdd_type;
		int io_port;

		char  reserved[9];
	};
#pragma pack()

	void reset_igram();
	void reset_membank();
	void reset_tracecounter();

public:
	MEMORY(VM* parent_vm, EMU* parent_emu, const char* identifier) : DEVICE(parent_vm, parent_emu, identifier)
	{
		set_class_name("MEMORY");
	}
	~MEMORY() {}

	// common functions
	void SET_BANK(int s, int e, uint8* w, uint8* r);

	void initialize();
	void reset();
	void write_data8(uint32 addr, uint32 data);
	uint32 read_data8(uint32 addr);
	void write_data16(uint32 addr, uint32 data) {
		write_data8(addr, data & 0xff); write_data8(addr + 1, data >> 8);
	}
	uint32 read_data16(uint32 addr) {
		return read_data8(addr) | (read_data8(addr + 1) << 8);
	}
	void write_io8(uint32 addr, uint32 data);
	void write_signal(int id, uint32 data, uint32 mask);

	// unique function
	void set_context_cpu(DEVICE* device) {
		d_cpu = device;
	}
	void set_context_pia(DEVICE* device) {
		d_pia = device;
	}
	void set_context_crtc(DEVICE* device) {
		d_crtc = device;
	}
	void set_context_display(DEVICE* device) {
		d_disp = device;
	}
	void set_context_acia(DEVICE* device) {
		d_acia = device;
	}
	void set_context_key(DEVICE* device) {
		d_key = device;
	}
	void set_context_sound(DEVICE* device) {
		d_sound = device;
	}
	void set_context_psg(DEVICE* device) {
		d_psg = device;
	}
	void set_context_psg9(DEVICE* device) {
		d_psg9 = device;
	}
	void set_context_comm(DEVICE* device) {
		d_comm = device;
	}
	void set_context_cmt(DEVICE* device) {
		d_cmt = device;
	}
	void set_context_timer(DEVICE* device) {
		d_timer = device;
	}
	void set_context_kanji(DEVICE* device) {
		d_kanji = device;
	}
	void set_context_fdc(DEVICE* device3, DEVICE* device5) {
		d_fdc3 = device3;
		d_fdc5 = device5;
	}
	void set_context_fdd(DEVICE* device) {
		d_fdd = device;
	}
	void set_context_pia_ex(DEVICE* device) {
		d_pia_ex = device;
	}
	void set_context_acia_ex(DEVICE* device) {
		d_acia_ex = device;
	}
	void set_context_comm1(DEVICE* device) {
		d_comm1 = device;
	}
	uint8* get_vram();

	uint8* get_color_ram() {
		return color_ram;
	}
	uint8* get_ig_ram() {
		return ig_ram;
	}

	void set_trace_counter();
	void change_memory_bank();

	void event_callback(int event_id, int err);
	void update_config();

	void save_state(FILEIO *fio);
	void load_state(FILEIO *fio);

};

#endif
